#!/usr/bin/env python

"""This module contains functions that can be used to setup the hybrid
structures and topologies needed for alchemical free energy calculations.
"""

import os
import sys
from copy import deepcopy
import numpy as np
from . import library
from .atom import Atom
from .model import Model
from .utils import mtpError, UnknownResidueError, MissingTopolParamError
from .geometry import Rotation, nuc_super, bb_super
from .mutdb import read_mtp_entry
from .forcefield import Topology, _check_case, _atoms_morphe
from .utils import get_mtp_file

__all__ = ['mutate', 'gen_hybrid_top', 'write_split_top', 'AbsRestraints']


# ==============
# Main Functions
# ==============
def mutate(m, mut_resid, mut_resname, ff, mut_chain=None,
           refB=None, inplace=False, verbose=False):
    """Creates an hybrid structure file.

    Parameters
    ----------
    m : Model
        the model to be mutated. See :py:class:`pmx.model.Model`.
    mut_resid : int
        the ID of the residue to be mutated.
    mut_resname : str
        the name of the target residue.
    ff : str
        the forcefield to use.
    mut_chain : str, optional
        the chain ID for the residue you want to mutate. This is needed if you
        have multiple chains and are providing a Model that did not have its
        residues renumbered.
    refB : str, optional
        reference structure file of the B state in PDB or GRO format. If it is
        provided, the dummy atoms will be built based on the position of the
        atoms in this structure. This option is available only for Protein
        mutations.
    inplace : bool, optional
        whether to modify the input Model in place. Default is False.
    verbose : bool, optional
        whether to print info. Default is False.

    Returns
    -------
    m2 : Model
        model with mutated residues. If argument 'inplace' is True, nothing is
        returned.
    """

    if inplace is True:
        m2 = m
    elif inplace is False:
        m2 = deepcopy(m)

    # convert Model to Angstrom units: this code works on A units
    if m2.unity == 'nm':
        m2.nm2a()
        if verbose:
            print('Model units have been converted from nm to A.')

    # get the residue based on the index
    # fetch_residue also checks the selection is valid and unique
    residue = m2.fetch_residue(idx=mut_resid, chain=mut_chain)
    # get the correct mtp file
    mtp_file = get_mtp_file(residue, ff)

    # Mutation if Protein
    if residue.moltype == 'protein':
        new_aa_name = _convert_aa_name(mut_resname)
        apply_aa_mutation(m=m2, residue=residue, new_aa_name=new_aa_name,
                          mtp_file=mtp_file, refB=refB, verbose=verbose)
    # Mutation if DNA or RNA
    elif residue.moltype in ['dna', 'rna']:
        new_nuc_name = mut_resname.upper()
        apply_nuc_mutation(m=m2, residue=residue, new_nuc_name=new_nuc_name,
                           mtp_file=mtp_file, verbose=verbose)

    if inplace is False:
        return m2


def apply_aa_mutation(m, residue, new_aa_name, mtp_file, refB=None,
                      verbose=False):

    if residue.resname == 'ILE':
        _rename_ile(residue)
    olkey = _convert_aa_name(residue.resname)

    # olkey should contain the correct one letter name of the WT residue
    # however, due to the different namings of the residues in the FFs
    # Lys needs to be checked once again: in OPLS Lys is non-protonated,
    # while in the other FFs it is protonated
    if ('opls' in mtp_file) and ('LYS' in residue.resname):
        olkey = _check_OPLS_LYS(residue)

    hybrid_residue_name = olkey+'2'+new_aa_name
    if verbose is True:
        print('log_> Residue to mutate: %s | %s | %s ' % (str(residue.id), residue.resname, residue.chain_id))
        print('log_> Mutation to apply: %s->%s' % (olkey, new_aa_name))
        print('log_> Hybrid residue name: %s' % hybrid_residue_name)
    hybrid_res, bonds, imps, diheds, rotdic = _get_hybrid_residue(residue_name=hybrid_residue_name,
                                                                  mtp_file=mtp_file,
                                                                  version='new',
                                                                  verbose=False)
    bb_super(residue, hybrid_res)

    # rename residue atoms
    hash1 = _rename_to_match_library(residue)
    hash2 = _rename_to_match_library(hybrid_res)
    _set_conformation(residue, hybrid_res, rotdic)
    if refB is not None:
        if verbose is True:
            print("log_> Set Bstate geometry according to the provided structure")
        mB = Model(refB, bPDBTER=True, rename_atoms=True, scale_coords='A')
        residueB = mB.residues[residue.id-1]
        bb_super(residue, residueB)
        for atom in hybrid_res.atoms:
            if (atom.name[0] == 'D') or atom.name.startswith('HV'):
                for atomB in residueB.atoms:
                    if atomB.name == hybrid_res.morphes[atom.name]['n1']:
                        atom.x = atomB.x
    _rename_back(residue, hash1)
    _rename_back(hybrid_res, hash2)

    # do the actual replacement
    m.replace_residue(residue=residue, new=hybrid_res, bKeepResNum=True)

    if verbose is True:
        print('log_> Inserted hybrid residue %s at position %s (chain %s)' %
              (hybrid_res.resname, str(hybrid_res.id), hybrid_res.chain_id))


def apply_nuc_mutation(m, residue, new_nuc_name, mtp_file, verbose=False):

    hybrid_residue_name, resname1, resname2 = get_nuc_hybrid_resname(residue, new_nuc_name)
    if verbose is True:
        print('log_> Residue to mutate: %d | %s | %s ' % (residue.id, residue.resname, residue.chain_id))
        print('log_> Mutation to apply: %s->%s' % (residue.resname[1], new_nuc_name))
        print('log_> Hybrid residue name: %s' % hybrid_residue_name)
    hybrid_res, bonds, imps, diheds, rotdic = _get_hybrid_residue(residue_name=hybrid_residue_name,
                                                                  mtp_file=mtp_file,
                                                                  version='new',
                                                                  verbose=False)

    nuc_super(residue, hybrid_res, resname1, resname2)
    for atom in hybrid_res.atoms:
        if atom.name[0] != 'D':
            atom.x = residue[atom.name].x
    m.replace_residue(residue=residue, new=hybrid_res, bKeepResNum=True)
    if verbose is True:
        print('log_> Inserted hybrid residue %s at position %d (chain %s)' %
              (hybrid_res.resname, hybrid_res.id, hybrid_res.chain_id))


def get_nuc_hybrid_resname(residue, new_nuc_name):

    if residue.moltype == 'dna':
        firstLetter = 'D'
    if residue.moltype == 'rna':
        firstLetter = 'R'

    # identify if the nucleotide is terminal
    for a in residue.atoms:
        if a.name == 'H3T':
            r1 = firstLetter+residue.resname[1]+'3'
            r2 = firstLetter+new_nuc_name+'3'
            dict_key = r1+'_'+r2
            if residue.moltype == 'rna':
                hybrid_residue_name = _rna_names[dict_key]
            else:
                hybrid_residue_name = _dna_names[dict_key]
            return(hybrid_residue_name, residue.resname[1], new_nuc_name)
        elif a.name == 'H5T':
            r1 = firstLetter+residue.resname[1]+'5'
            r2 = firstLetter+new_nuc_name+'5'
            dict_key = r1+'_'+r2
            if residue.moltype == 'rna':
                hybrid_residue_name = _rna_names[dict_key]
            else:
                hybrid_residue_name = _dna_names[dict_key]
            return(hybrid_residue_name, residue.resname[1], new_nuc_name)
    hybrid_residue_name = residue.resname+new_nuc_name
    return(hybrid_residue_name, residue.resname[1], new_nuc_name)


def gen_hybrid_top(topol, recursive=True, verbose=False, scaleDih=1.0):
    """Fills the bstate of a topology file containing pmx hybrid residues. This
    can be either a top or itp file. If the file contains other itp files via
    include statements, the function can iterate through them if the recursive
    flag is set to True.

    It returns a Topology instance with b states filled for the input topol
    instance, and optionally a list of additional Topology instances for the
    included itp files. If there are no itp files, or if recursive is set to
    False, then an empty list is returned.

    Parameters
    ----------
    topol : Topology
        input Topology instance
    recursive : bool, optional
        whether to fill b states also for all itp files included in the
        topology file. Default is True.
    verbose : bool
        whether to print out info. Default is False.
    scaleDih : float
        scaling for dihedrals with a dummy.

    Returns
    -------
    top, itps : obj, list of obj
        output Topology instance with b states assigned (top), and
        list of Topology instances for the additional itp files (itps)
        included in the input topology.
    """

    # ff setup
    ff = topol.forcefield
    ffbonded_file = os.path.join(topol.ffpath, 'ffbonded.itp')

    # main function that returns the Topology with filled B states
    def process_topol(topol, ff, ffbonded_file, verbose=False, scaleDih=1.0):
        pmxtop = deepcopy(topol)
        # create model with residue list
        m = Model(atoms=pmxtop.atoms, renumber_residues=False)
        # need to sequentially renumber residues, 
        # because --keep_resid flag may leave residues numbered non-sequentially
        counter = 0
        for r in m.residues:
            counter += 1 
            r.id = counter
        # use model residue list
        pmxtop.residues = m.residues
        # get list of hybrid residues and their params
        rlist, rdic = _get_hybrid_residues(m=m, ff=ff, version='new',
                                           verbose=verbose)
        # correct b-states
        pmxtop.assign_fftypes()
        if verbose is True:
            for r in rlist:
                print('log_> Hybrid Residue -> %d | %s ' % (r.id, r.resname))

        _find_bonded_entries(pmxtop, verbose=verbose)
        _find_angle_entries(pmxtop, verbose=verbose)
        dih_predef_default = []
        _find_predefined_dihedrals(pmxtop, rlist, rdic, ffbonded_file,
                                   dih_predef_default, ff, scaleDih=scaleDih)
        _find_dihedral_entries(pmxtop, rlist, rdic, dih_predef_default,
                               verbose=verbose, scaleDih=scaleDih)

        _add_extra_DNA_RNA_impropers(pmxtop, rlist, 1, [180, 40, 2], [180, 40, 2])

        if verbose is True:
            print('log_> Total charge of state A = %.f' % pmxtop.get_qA())
            print('log_> Total charge of state B = %.f' % pmxtop.get_qB())

        # if prolines are involved, break one bond (CD-CG)
        # and angles X-CD-CG, CD-CG-X
        # and dihedrals with CD and CG
        _proline_decouplings(pmxtop, rlist, rdic)
        # also decouple all dihedrals with [CD and N] and [CB and Calpha] for proline
        _proline_dihedral_decouplings(pmxtop, rlist, rdic)
        return pmxtop

    # -------------
    # Main topology
    # -------------
    if verbose is True:
        print('\nlog_> Reading input %s file "%s"'
              % (topol.filename.split('.')[-1], topol.filename))
    pmx_top = process_topol(topol=topol, ff=ff, ffbonded_file=ffbonded_file,
                            verbose=verbose, scaleDih=scaleDih)

    # ------------------------------------
    # itps too if asked for and if present
    # ------------------------------------
    itps = topol.include_itps
    pmx_itps = []
    if recursive is True and len(itps) > 0:
        for itp, where in itps:
            if verbose is True:
                print('\nlog_> Reading input itp file "%s""' % (itp))
            itp_path = os.path.dirname(os.path.relpath(topol.filename))
            itp_filename = os.path.join(itp_path, itp)
            topol2 = Topology(itp_filename, ff=ff, version='new')
            # fill b states
            pmx_itp = process_topol(topol=topol2, ff=ff,
                                    ffbonded_file=ffbonded_file,
                                    verbose=verbose)
            pmx_itps.append(pmx_itp)

    return pmx_top, pmx_itps


def write_split_top(pmxtop, outfile='pmxtop.top', scale_mass=False,
                    verbose=False):
    """
    Write three topology files to be used for three separate free energy
    calculations: charges off, vdw on, changes on. This can be useful when one
    wants to avoid using a soft-core for the electrostatic interactions.

    Parameters
    ----------
    pmxtop : Topology
        Topology object with B states defined.
    outfile : str
        name of output topology file.
    scale_mass : bool
        whether to scale the masses of morphing atoms
    verbose : bool
        whether to print information or not

    Returns
    -------
    None
    """

    root, ext = os.path.splitext(outfile)
    out_file_qoff = root + '_qoff' + ext
    out_file_vdw = root + '_vdw' + ext
    out_file_qon = root + '_qon' + ext

    # get charge of states A/B for hybrid residues
    qA = pmxtop.get_hybrid_qA()
    qB = pmxtop.get_hybrid_qB()

    if verbose is True:
        print('------------------------------------------------------')
        print('log_> Creating splitted topologies............')

        print('log_> Making "qoff" topology : "%s"' % out_file_qoff)
    contQ = deepcopy(qA)
    pmxtop.write(out_file_qoff, stateQ='AB', stateTypes='AA', dummy_qB='off',
                 scale_mass=scale_mass, target_qB=qA, stateBonded='AA',
                 full_morphe=False)
    if verbose is True:
        print('log_> Charge of state A: %g' % pmxtop.qA)
        print('log_> Charge of state B: %g' % pmxtop.qB)
    if verbose is True:
        print('------------------------------------------------------')
        print('log_> Making "vdw" topology : "%s"' % out_file_vdw)
    contQ = deepcopy(qA)
    pmxtop.write(out_file_vdw, stateQ='BB', stateTypes='AB', dummy_qA='off',
                 dummy_qB='off', scale_mass=scale_mass,
                 target_qB=contQ, stateBonded='AB', full_morphe=False)
    if verbose is True:
        print('log_> Charge of state A: %g' % pmxtop.qA)
        print('log_> Charge of state B: %g' % pmxtop.qB)
        print('------------------------------------------------------')

        print('log_> Making "qon" topology : "%s"' % out_file_qon)
    pmxtop.write(out_file_qon, stateQ='BB', stateTypes='BB', dummy_qA='off',
                 dummy_qB='on', scale_mass=scale_mass, target_qB=qB,
                 stateBonded='BB', full_morphe=False)
    if verbose is True:
        print('log_> Charge of state A: %g' % pmxtop.qA)
        print('log_> Charge of state B: %g' % pmxtop.qB)
        print('------------------------------------------------------')


def decouple_mol(top):
    """Setup topology for decoupling a small molecule. The nonbonded
    interactions are on in state A and are off in state B.

    Params
    ------
    top : Topology
        topology object that will be modified

    Returns
    -------
    None
    """

    for atom in top.atoms:
        atom.qB = 0.0
        atom.mB = atom.m
        atom.typeB = 'dum'
        atom.atomtypeB = 'dum'


    # create the dummy atomtype
    dummy = {}
    dummy['name'] = 'dum'
    dummy['bond_type'] = 'dum'
    dummy['mass'] = 0.0
    dummy['charge'] = 0.0
    dummy['ptype'] = 'A'
    dummy['sigma'] = 0.0
    dummy['epsilon'] = 0.0

    # add the dummy to the topology atomtypes
    top.atomtypes.append(dummy)


def couple_mol(top):
    """Setup topology for coupling a small molecule. The nonbonded
    interactions are off in state A and are on in state B.

    Params
    ------
    top : Topology
        topology object that will be modified

    Returns
    -------
    None
    """

    for atom in top.atoms:
        atom.qB = atom.q
        atom.q = 0.0
        atom.mB = atom.m
        atom.typeB = atom.type
        atom.atomtypeB = atom.type
        atom.type = 'dum'
        atom.atomtype = 'dum'

    # create the dummy atomtype
    dummy = {}
    dummy['name'] = 'dum'
    dummy['bond_type'] = 'dum'
    dummy['mass'] = 0.0
    dummy['charge'] = 0.0
    dummy['ptype'] = 'A'
    dummy['sigma'] = 0.0
    dummy['epsilon'] = 0.0

    # add the dummy to the topology atomtypes
    top.atomtypes.append(dummy)


# ===============
# HelperFunctions
# ===============
def _scale_dih( d, scale ):
    # scale dihedral
    dih = deepcopy(d)
    if dih[0]==9 or dih[0]==4 or dih[0]==1 or dih[0]==2:
        dih[2] *= scale
    elif dih[0]==3:
        dih[1] *= scale
        dih[2] *= scale
        dih[3] *= scale
        dih[4] *= scale
        dih[5] *= scale
        dih[6] *= scale
    return(dih)

def _check_OPLS_LYS(res):
    if res.has_atom('HZ3'):
        return('K')
    else:
        return('O')


def _convert_aa_name(aa):
    # firstly, some special deprotonated cases
    if aa == 'CM':
        return(aa.upper())
    elif len(aa) == 1:
        return aa.upper()
    elif len(aa) in [3, 4]:
        return library._ext_one_letter[aa.upper()]
    else:
        raise UnknownResidueError(aa)


def _rename_ile(residue):
    dic = {'CD':  'CD1',
           'HD1': 'HD11',
           'HD2': 'HD12',
           'HD3': 'HD13'
           }
    for key, value in dic.items():
        try:
            atom = residue[key]
            atom.name = value
        except:
            pass


def _rename_to_match_library(res):
    name_hash = {}
    atoms = res.atoms
    for atom in atoms:
        foo = atom.name
        # for serine
        if (atom.resname == 'SER') and (atom.name == 'HG1'):
            atom.name = 'HG'
        if ('S2' in atom.resname) and (atom.name == 'HG1'):
            atom.name = 'HG'
        # for cysteine
        if (atom.resname == 'CYS') and (atom.name == 'HG1'):
            atom.name = 'HG'
        if ('C2' in atom.resname) and (atom.name == 'HG1'):
            atom.name = 'HG'

        name_hash[atom.name] = foo
    return name_hash


def _rename_back(res, name_hash):
    for atom in res.atoms:
        atom.name = name_hash[atom.name]


def _set_conformation(old_res, new_res, rotdic):
    old_res.get_real_resname()
    dihedrals = library._aa_dihedrals[old_res.real_resname]
    for key, lst in rotdic.items():
        new = new_res.fetchm(lst)
        rotdic[key] = new
    chis = []
    for key in rotdic.keys():
        at1, at2 = key.split('-')
        for d in dihedrals:
            if d[1] == at1 and d[2] == at2 \
                   and d[-1] != -1:
                chis.append(d)
    for d in chis:
        atoms = old_res.fetchm(d[:4])
        phi = atoms[0].dihedral(atoms[1], atoms[2], atoms[3])
        atoms2 = new_res.fetchm(d[:4])
        phi2 = atoms2[0].dihedral(atoms2[1], atoms2[2], atoms2[3])
        diff = phi-phi2
        a1, a2 = new_res.fetchm(d[1:3])
        key = a1.name+'-'+a2.name
        atoms = rotdic[key]
        rot = Rotation(a1.x, a2.x)
        for atom in atoms:
            atom.x = rot.apply(atom.x, diff)
    for atom in new_res.atoms:
        if (atom.name[0] != 'D') and (not atom.name.startswith('HV')):
            atom.x = old_res[atom.name].x


def _proline_dihedral_decouplings(topol, rlist, rdic):
    for r in rlist:
        if 'P' in r.resname:
            # identify in which state proline sits
            prolineState = 'A'
            if '2P' in r.resname:
                prolineState = 'B'

            # find the atoms for which the bond needs to be broken
            prolineCDatom = ''
            prolineCBatom = ''
            prolineNatom = ''
            prolineCAatom = ''
            for a in r.atoms:
                if (prolineState == 'A') and (a.name == 'CD'):
                    prolineCDatom = a
                elif (prolineState == 'A') and (a.name == 'N'):
                    prolineNatom = a
                elif (prolineState == 'A') and (a.name == 'CB'):
                    prolineCBatom = a
                elif (prolineState == 'A') and (a.name == 'CA'):
                    prolineCAatom = a
                elif (prolineState == 'B') and (a.name == 'DCD'):
                    prolineCDatom = a
                elif (prolineState == 'B') and (a.name == 'N'):
                    prolineNatom = a
                elif (prolineState == 'B') and (a.name == 'DCB'):
                    prolineCBatom = a
                elif (prolineState == 'B') and (a.name == 'CA'):
                    prolineCAatom = a

            # decouple [CD and N] dihedrals
            for dih in topol.dihedrals:
                a1 = dih[0]
                a2 = dih[1]
                a3 = dih[2]
                a4 = dih[3]
                func = dih[4]
                if ((a1.id == prolineCDatom.id and a2.id == prolineNatom.id)) or \
                   ((a2.id == prolineCDatom.id and a1.id == prolineNatom.id)) or \
                   ((a1.id == prolineCDatom.id and a3.id == prolineNatom.id)) or \
                   ((a3.id == prolineCDatom.id and a1.id == prolineNatom.id)) or \
                   ((a1.id == prolineCDatom.id and a4.id == prolineNatom.id)) or \
                   ((a4.id == prolineCDatom.id and a1.id == prolineNatom.id)) or \
                   ((a2.id == prolineCDatom.id and a3.id == prolineNatom.id)) or \
                   ((a3.id == prolineCDatom.id and a2.id == prolineNatom.id)) or \
                   ((a2.id == prolineCDatom.id and a4.id == prolineNatom.id)) or \
                   ((a4.id == prolineCDatom.id and a2.id == prolineNatom.id)) or \
                   ((a3.id == prolineCDatom.id and a4.id == prolineNatom.id)) or \
                   ((a4.id == prolineCDatom.id and a3.id == prolineNatom.id)):
                    if (dih[5] == 'NULL') or (dih[6] == 'NULL'):
                        continue
                    if prolineState == 'A':
                        if func == 3:
                            dih[6] = [func, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
                        elif func == 2:
                            dih[6] = [func, dih[6][1], 0.0]
                        else:
                            dih[6] = [func, dih[6][1], 0.0, dih[6][-1]]
                    if prolineState == 'B':
                        if func == 3:
                            dih[5] = [func, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
                        elif func == 2:
                            dih[5] = [func, dih[6][1], 0.0]
                        else:
                            dih[5] = [func, dih[6][1], 0.0, dih[6][-1]]

            # decouple [CB and CA] dihedrals
            for dih in topol.dihedrals:
                a1 = dih[0]
                a2 = dih[1]
                a3 = dih[2]
                a4 = dih[3]
                func = dih[4]
                if ((a1.id == prolineCBatom.id and a2.id == prolineCAatom.id)) or \
                   ((a2.id == prolineCBatom.id and a1.id == prolineCAatom.id)) or \
                   ((a1.id == prolineCBatom.id and a3.id == prolineCAatom.id)) or \
                   ((a3.id == prolineCBatom.id and a1.id == prolineCAatom.id)) or \
                   ((a1.id == prolineCBatom.id and a4.id == prolineCAatom.id)) or \
                   ((a4.id == prolineCBatom.id and a1.id == prolineCAatom.id)) or \
                   ((a2.id == prolineCBatom.id and a3.id == prolineCAatom.id)) or \
                   ((a3.id == prolineCBatom.id and a2.id == prolineCAatom.id)) or \
                   ((a2.id == prolineCBatom.id and a4.id == prolineCAatom.id)) or \
                   ((a4.id == prolineCBatom.id and a2.id == prolineCAatom.id)) or \
                   ((a3.id == prolineCBatom.id and a4.id == prolineCAatom.id)) or \
                   ((a4.id == prolineCBatom.id and a3.id == prolineCAatom.id)):
                    if (dih[5] == 'NULL') or (dih[6] == 'NULL'):
                        continue
                    if prolineState == 'A':
                        if func == 3:
                            dih[6] = [func, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
                        elif func == 2:
                            dih[6] = [func, dih[6][1], 0.0]
                        else:
                            dih[6] = [func, dih[6][1], 0.0, dih[6][-1]]
                    if prolineState == 'B':
                        if func == 3:
                            dih[5] = [func, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
                        elif func == 2:
                            dih[5] = [func, dih[6][1], 0.0]
                        else:
                            dih[5] = [func, dih[6][1], 0.0, dih[6][-1]]


def _proline_decouplings(topol, rlist, rdic):
    for r in rlist:
        if 'P' in r.resname:
            # identify in which state proline sits
            prolineState = 'A'
            if '2P' in r.resname:
                prolineState = 'B'

            # find the atoms for which the bond needs to be broken
            prolineCDatom = ''
            prolineCGatom = ''
            for a in r.atoms:
                if (prolineState == 'A') and (a.name == 'CD'):
                    prolineCDatom = a
                elif (prolineState == 'A') and (a.name == 'CG'):
                    prolineCGatom = a
                elif (prolineState == 'B') and (a.name == 'DCD'):
                    prolineCDatom = a
                elif (prolineState == 'B') and (a.name == 'DCG'):
                    prolineCGatom = a

            # break the bond
            for b in topol.bonds:
                a1 = b[0]
                a2 = b[1]
                if ((a1.id == prolineCDatom.id) and (a2.id == prolineCGatom.id)) or \
                   ((a2.id == prolineCDatom.id) and (a1.id == prolineCGatom.id)):
                    if prolineState == 'A':
                        b[4] = deepcopy(b[3])
                        b[4][2] = 0.0
                    elif prolineState == 'B':
                        b[3] = deepcopy(b[4])
                        b[3][2] = 0.0

            # decouple angles
            for ang in topol.angles:
                a1 = ang[0]
                a2 = ang[1]
                a3 = ang[2]
                func = ang[3]
                if ((a1.id == prolineCDatom.id and a2.id == prolineCGatom.id)) or \
                   ((a2.id == prolineCDatom.id and a1.id == prolineCGatom.id)) or \
                   ((a1.id == prolineCDatom.id and a3.id == prolineCGatom.id)) or \
                   ((a3.id == prolineCDatom.id and a1.id == prolineCGatom.id)) or \
                   ((a2.id == prolineCDatom.id and a3.id == prolineCGatom.id)) or \
                   ((a3.id == prolineCDatom.id and a2.id == prolineCGatom.id)):
                    if prolineState == 'A':
                        if func == 5:
                            ang[5] = [1, 0, 0, 0, 0]
                        else:
                            ang[5] = [1, 0, 0]
                    elif prolineState == 'B':
                        if func == 5:
                            ang[4] = [1, 0, 0, 0, 0]
                        else:
                            ang[4] = [1, 0, 0]

            # decouple dihedrals
            for dih in topol.dihedrals:
                a1 = dih[0]
                a2 = dih[1]
                a3 = dih[2]
                a4 = dih[3]
                func = dih[4]
                if ((a1.id == prolineCDatom.id and a2.id == prolineCGatom.id)) or \
                   ((a2.id == prolineCDatom.id and a1.id == prolineCGatom.id)) or \
                   ((a1.id == prolineCDatom.id and a3.id == prolineCGatom.id)) or \
                   ((a3.id == prolineCDatom.id and a1.id == prolineCGatom.id)) or \
                   ((a1.id == prolineCDatom.id and a4.id == prolineCGatom.id)) or \
                   ((a4.id == prolineCDatom.id and a1.id == prolineCGatom.id)) or \
                   ((a2.id == prolineCDatom.id and a3.id == prolineCGatom.id)) or \
                   ((a3.id == prolineCDatom.id and a2.id == prolineCGatom.id)) or \
                   ((a2.id == prolineCDatom.id and a4.id == prolineCGatom.id)) or \
                   ((a4.id == prolineCDatom.id and a2.id == prolineCGatom.id)) or \
                   ((a3.id == prolineCDatom.id and a4.id == prolineCGatom.id)) or \
                   ((a4.id == prolineCDatom.id and a3.id == prolineCGatom.id)):
                    if (dih[5] == 'NULL') or (dih[6] == 'NULL'):
                        continue

                    if prolineState == 'A':
                        if func == 3:
                            dih[6] = [func, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
                        elif func == 2:
                            dih[6] = [func, dih[6][1], 0.0]
                        else:
                            dih[6] = [func, dih[6][1], 0.0, dih[6][-1]]
                    if prolineState == 'B':
                        if func == 3:
                            dih[5] = [func, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
                        elif func == 2:
                            dih[5] = [func, dih[6][1], 0.0]
                        else:
                            dih[5] = [func, dih[6][1], 0.0, dih[6][-1]]


def _find_bonded_entries(topol, verbose=False):
    count = 0
    for b in topol.bonds:
        try:
            a1, a2, func = b
        except: # for the cases where params explicitly listed
            a1, a2, func, otherParam = b
        A, B = _check_case([a1, a2])
        if a1.atomtypeB is not None or a2.atomtypeB is not None:
            count += 1
            error_occured = False
            astate = None
            bstate = None
            if A == 'AA' and B == 'AA':  # we need A and B state
                astate = topol.BondedParams.get_bond_param(a1.type, a2.type)
                bstate = topol.BondedParams.get_bond_param(a1.typeB, a2.typeB)
                b.extend([astate, bstate])
            elif 'D' in A and B == 'AA':
                bstate = topol.BondedParams.get_bond_param(a1.typeB, a2.typeB)
                astate = bstate
                b.extend([astate, bstate])
            elif 'D' in B and A == 'AA':
                astate = topol.BondedParams.get_bond_param(a1.type, a2.type)
                bstate = astate
                b.extend([astate, bstate])
            # catch errors
            elif 'D' in B and 'D' in A:
                print('Error: fake bond '
                      '%s-%s (%s-%s -> %s-%s)' % (a1.name, a2.name,
                                                  a1.atomtype, a2.atomtype,
                                                  a1.atomtypeB, a2.atomtypeB))
                sys.exit(1)
            if astate is None:
                print('Error A: No bond entry found for astate '
                      '%s-%s (%s-%s -> %s-%s)' % (a1.name, a2.name,
                                                  a1.atomtype, a2.atomtype,
                                                  a1.type, a2.type))
                print('Resi = %s' % a1.resname)
                error_occured = True

            if bstate is None:
                print('Error B: No bond entry found for bstate '
                      '%s-%s (%s-%s -> %s-%s)' % (a1.name, a2.name,
                                                  a1.atomtypeB, a2.atomtypeB,
                                                  a1.typeB, a2.typeB))
                error_occured = True
            if error_occured:
                sys.exit(1)
    if verbose is True:
        print('log_> Making bonds for state B -> %d bonds with perturbed atoms'
              % count)


def _find_angle_entries(topol, verbose=False):
    count = 0
    for a in topol.angles:
        try:
            a1, a2, a3, func = a
        except: # for the cases where params explicitly listed
            a1, a2, a3, func, otherParam = a
        astate = None
        bstate = None
        if a1.atomtypeB is not None or \
           a2.atomtypeB is not None or \
           a3.atomtypeB is not None:
            A, B = _check_case([a1, a2, a3])
            count += 1
            if 'D' in A and 'D' in B:  # fake angle
                if func == 5:
                    astate = [1, 0, 0, 0, 0]
                    bstate = [1, 0, 0, 0, 0]
                else:
                    astate = [1, 0, 0]
                    bstate = [1, 0, 0]
                a.extend([astate, bstate])
            elif A == 'AAA' and 'D' in B:   # take astate
                astate = topol.BondedParams.get_angle_param(a1.type,
                                                            a2.type,
                                                            a3.type)
                bstate = astate
                a.extend([astate, bstate])
            elif 'D' in A and B == 'AAA':   # take bstate
                bstate = topol.BondedParams.get_angle_param(a1.typeB,
                                                            a2.typeB,
                                                            a3.typeB)
                astate = bstate
                a.extend([astate, bstate])
            elif A == 'AAA' and B == 'AAA':
                if a1.atomtypeB != a1.atomtype or \
                   a2.atomtypeB != a2.atomtype or \
                   a3.atomtypeB != a3.atomtype:
                    astate = topol.BondedParams.get_angle_param(a1.type,
                                                                a2.type,
                                                                a3.type)
                    bstate = topol.BondedParams.get_angle_param(a1.typeB,
                                                                a2.typeB,
                                                                a3.typeB)
                    a.extend([astate, bstate])
                else:
                    astate = topol.BondedParams.get_angle_param(a1.type,
                                                                a2.type,
                                                                a3.type)
                    bstate = astate
                    a.extend([astate, bstate])
            if astate is None:
                raise MissingTopolParamError("No angle entry (state A)",
                                             [a1, a2, a3])
            if bstate is None:
                raise MissingTopolParamError("No angle entry (state B)",
                                             [a1, a2, a3])

    if verbose is True:
        print('log_> Making angles for state B -> %d angles with perturbed atoms'
              % count)


def _find_dihedral_entries(topol, rlist, rdic, dih_predef_default,
                           verbose=False, scaleDih=1.0):
    count = 0
    nfake = 0
    # here I will accumulate multiple entries of type 9 dihedrals
    dih9 = []
    # need to store already visited angles to avoid multiple re-definitions
    visited_dih = []
    # scale dihedrals with dummies
#    scaleDih = 0.0

    for d in topol.dihedrals:
        if len(d) >= 6:
            # only consider the dihedral, if it has not been encountered so far
            undef = 0
            encountered = 0

            undef, encountered = _is_dih_undef(dih_predef_default, d)
            encountered = _is_dih_encountered_strict(visited_dih, d,
                                                     encountered)
            if(encountered == 1):
                continue

            visited_dih.append(d)

            if undef == 3 or undef == 4:
                a1 = d[0]
                a2 = d[1]
                a3 = d[2]
                a4 = d[3]
                val = ''
                func = 9
            elif(len(d) == 6):
                a1, a2, a3, a4, func, val = d
            else:
                a1, a2, a3, a4, func, val, rest = d

            backup_d = d[:6]

            if _atoms_morphe([a1, a2, a3, a4]):
                A, B = _check_case(d[:4])
                if A != 'AAAA' and B != 'AAAA':
                    nfake += 1
                    # these are fake dihedrals
                    d[5] = 'NULL'
                    d.append('NULL')
                else:
                    count += 1
                    astate = []
                    bstate = []
                    if A == 'AAAA' and B != 'AAAA':
                        foo = topol.BondedParams.get_dihedral_param(a1.type,
                                                                    a2.type,
                                                                    a3.type,
                                                                    a4.type,
                                                                    func)
                        # need to check if the dihedral has torsion pre-defined
                        counter = _check_dih_ILDN_OPLS(topol, rlist, rdic,
                                                       a1, a2, a3, a4)
                        if counter == 42:
                            continue

                        for ast in foo:
                            if counter == 0:
                                d[5] = ast
                                bst = _scale_dih( ast, scaleDih )
                                d.append(bst)
                            else:
                                alus = backup_d[:]
                                alus[5] = ast
                                bst = _scale_dih( ast, scaleDih )
                                alus.append(bst)
                                dih9.append(alus)
                            counter = 1

                    elif B == 'AAAA' and A != 'AAAA':
                        foo = topol.BondedParams.get_dihedral_param(a1.typeB,
                                                                    a2.typeB,
                                                                    a3.typeB,
                                                                    a4.typeB,
                                                                    func)
                        # need to check if the dihedral has torsion pre-defined
                        counter = _check_dih_ILDN_OPLS(topol, rlist, rdic, a1, a2, a3, a4)

                        if counter == 42:
                            continue

                        for bst in foo:
                            if counter == 0:
                                ast = _scale_dih( bst, scaleDih )
                                d[5] = ast
                                d.append(bst)
                            else:
                                alus = backup_d[:]
                                ast = _scale_dih( bst, scaleDih )
                                alus[5] = ast
                                alus.append(bst)
                                dih9.append(alus)
                            counter = 1

                    elif A == 'AAAA' and B == 'AAAA':
                        # disappear/appear dihedrals, do not morphe
                        if val == '':
                            if(undef != 4):
                                astate = topol.BondedParams.get_dihedral_param(a1.type,
                                                                               a2.type,
                                                                               a3.type,
                                                                               a4.type,
                                                                               func)
                        else:
                            if(undef != 4):
                                astate = topol.BondedParams.get_dihedral_param(a1.type,
                                                                               a2.type,
                                                                               a3.type,
                                                                               a4.type,
                                                                               func)
                        # if types_morphe([a1,a2,a3,a4]):
                        if (undef != 3):
                                bstate = topol.BondedParams.get_dihedral_param(a1.typeB,
                                                                               a2.typeB,
                                                                               a3.typeB,
                                                                               a4.typeB,
                                                                               func)

                        if undef == 1 and astate == []:
                            continue
                        elif undef == 1 and (astate[0][0] == 4 or astate[0][0] == 2):
                            continue
                        elif undef == 2 and bstate == []:
                            continue
                        elif undef == 2 and (bstate[0][0] == 4 or bstate[0][0] == 2):
                            continue

                        # need to check if the dihedral has torsion pre-defined
                        counter = _check_dih_ILDN_OPLS(topol, rlist, rdic,
                                                       a1, a2, a3, a4)

                        # torsion for both states defined, change nothing
                        if counter == 42:
                            continue

                        # A state disappears (when going to B)
                        for ast in astate:
                            if counter == 0:
                                d[5] = ast
                                if ast[0] == 3:
                                    bst = [ast[0], 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
                                elif ast[0] == 2:
                                    bst = [ast[0], ast[1], 0.0]
                                else:
                                    bst = [ast[0], ast[1], 0.0, ast[-1]]
                                if len(d) == 6:
                                    d.append(bst)
                                else:
                                    d[6] = bst
                                counter = 1
                            elif counter == 1 or counter == 3:
                                alus = backup_d[:]
                                alus[5] = ast
                                if ast[0] == 3:
                                    bst = [ast[0], 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
                                elif ast[0] == 2:
                                    bst = [ast[0], ast[1], 0.0]
                                else:
                                    bst = [ast[0], ast[1], 0.0, ast[-1]]

                                alus.append(bst)
                                dih9.append(alus)

                        # B state disappears (when going to A)
                        # all must go to the new (appendable) dihedrals
                        for bst in bstate:
                            if counter == 0:
                                if bst[0] == 3:
                                    ast = [bst[0], 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
                                elif bst[0] == 2:
                                    ast = [bst[0], bst[1], 0.0]
                                else:
                                    ast = [bst[0], bst[1], 0.0, bst[-1]]
                                d[5] = ast
                                if len(d) == 6:
                                    d.append(bst)
                                else:
                                    d[6] = bst
                                counter = 1
                            elif counter == 1 or counter == 2:
                                alus = backup_d[:]
                                if ast[0] == 3:
                                    ast = [bst[0], 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
                                elif ast[0] == 2:
                                    ast = [bst[0], bst[1], 0.0]
                                else:
                                    ast = [bst[0], bst[1], 0.0, bst[-1]]
                                alus[5] = ast
                                alus.append(bst)
                                dih9.append(alus)


                    if astate is None:
                        print('Error: No dihedral angle found (state A: predefined state B) for:')
                        print('{0} {1} {2} {3}'.format(a1.resname, a2.resname, a3.resname, a4.resname))
                        print('{0} {1} {2} {3}'.format(a1.name, a2.name, a3.name, a4.name, func))
                        print('{0} {1} {2} {3}'.format(a1.atomtype, a2.atomtype, a3.atomtype, a4.atomtype))
                        print('{0} {1} {2} {3}'.format(a1.type, a2.type, a3.type, a4.type))
                        print('{0} {1} {2} {3}'.format(a1.atomtypeB, a2.atomtypeB, a3.atomtypeB, a4.atomtypeB))
                        print('{0} {1} {2} {3}'.format(a1.typeB, a2.typeB, a3.typeB, a4.typeB))
                        print('{0}'.format(astate))
                        print('{0}'.format(d))
                        sys.exit(1)

                    if bstate is None:
                        print('Error: No dihedral angle found (state B: predefined state A) for:')
                        print('{0} {1} {2} {3}'.format(a1.resname, a2.resname, a3.resname, a4.resname))
                        print('{0} {1} {2} {3}'.format(a1.name, a2.name, a3.name, a4.name, func))
                        print('{0} {1} {2} {3}'.format(a1.atomtype, a2.atomtype, a3.atomtype, a4.atomtype))
                        print('{0} {1} {2} {3}'.format(a1.type, a2.type, a3.type, a4.type))
                        print('{0} {1} {2} {3}'.format(a1.atomtypeB, a2.atomtypeB, a3.atomtypeB, a4.atomtypeB))
                        print('{0} {1} {2} {3}'.format(a1.typeB, a2.typeB, a3.typeB, a4.typeB))
                        print('{0}'.format(d))
                        sys.exit(1)

                    # the previous two if sentences will kill the execution
                    # if anything goes wrong. Therefore, I am not afraid to do
                    # d.append() already in the previous steps


    topol.dihedrals.extend(dih9)
    if verbose is True:
        print('log_> Making dihedrals for state B -> %d dihedrals with perturbed atoms' % count)
        print('log_> Removed %d fake dihedrals' % nfake)


def _check_dih_ILDN_OPLS(topol, rlist, rdic, a1, a2, a3, a4):
    counter = 0
    for r in rlist:
        if r.id == a2.resnr:
            dih = rdic[r.resname][3]
            for d in dih:
                if counter == 1:
                    break
                al = []
                for name in d[:4]:
                    atom = r.fetch(name)[0]
                    al.append(atom)

                if (a1.id == al[0].id and a2.id == al[1].id and
                   a3.id == al[2].id and a4.id == al[3].id) or \
                   (a1.id == al[3].id and a2.id == al[2].id and
                   a3.id == al[1].id and a4.id == al[0].id):
                    # ---------------
                    # checks for OPLS
                    # ---------------
                    if d[4].startswith('dih_') and d[5].startswith('dih_'):
                        counter = 42
                        break
                    elif d[4].startswith('dih_') and 'un' in d[5]:
                        counter = 2
                        break
                    elif 'un' in d[4] and d[5].startswith('dih_'):
                        counter = 3
                        break
                    # ---------------
                    # checks for ILDN
                    # ---------------
                    if d[4].startswith('torsion') and d[5].startswith('torsion'):
                        counter = 42
                        break
                    elif d[4].startswith('torsion') and 'un' in d[5]:
                        counter = 2
                        break
                    elif 'un' in d[4] and d[5].startswith('torsion'):
                        counter = 3
                        break
                    elif 'un' in d[4]:
                        counter = 1
                        break
    return counter


def _is_dih_undef(visited_dih, d):
    encountered = 0
    undef = 0
    for dih in visited_dih:
        if (dih[0].id == d[0].id and dih[1].id == d[1].id
           and dih[2].id == d[2].id and dih[3].id == d[3].id):
            if (dih[-1] == 'undefA'):
                undef = 1
                break
            elif (dih[-1] == 'undefB'):
                undef = 2
                break
            elif (dih[-1] == 'undefA_ildn'):
                undef = 3
                break
            elif (dih[-1] == 'undefB_ildn'):
                undef = 4
                break
            else:
                encountered = 1
                break

    return (undef, encountered)


def _is_dih_encountered_strict(visited_dih, d, encountered):
    for dih in visited_dih:
        if (dih[0].id == d[0].id and dih[1].id == d[1].id
           and dih[2].id == d[2].id and dih[3].id == d[3].id):
            encountered = 1
            break
    return encountered


def _get_torsion_multiplicity(name):
    foo = list(name)
    mult = int(foo[-1])
    return mult


def _explicit_defined_dihedrals(filename, ff):
    """ get the #define dihedral entries explicitly for ILDN
    """
    lines = open(filename).readlines()
    output = {}
    ffnamelower = ff.lower()
    for line in lines:
        if line.startswith('#define'):
            entr = line.split()
            name = entr[1]
            if ffnamelower.startswith('amber'):
                params = [9, float(entr[2]), float(entr[3]), int(entr[4])]
            elif ffnamelower.startswith('opls'):
                if len(entr) == 8:  # dihedral
                    params = [3, float(entr[2]), float(entr[3]), float(entr[4]),
                              float(entr[5]), float(entr[6]), float(entr[7])]
                elif len(entr) == 5:
                    params = [1, float(entr[2]), float(entr[3]), float(entr[4])]
            output[name] = params
    return output


def _is_ildn_dih_encountered(ildn_used, d, encountered):
    for dih in ildn_used:
        if (dih[0] == d[0] and dih[1] == d[1] and dih[2] == d[2]
           and dih[3] == d[3] and dih[4] == d[4]):
            encountered = 1
    return encountered


def _find_predefined_dihedrals(topol, rlist, rdic, ffbonded,
                               dih_predef_default, ff, scaleDih=1.0):

    dih9 = []  # here I will accumulate multiple entries of type 9 dihedrals
    explicit_def = _explicit_defined_dihedrals(ffbonded, ff)
    ildn_used = []  # ildn dihedrals that already were encountered
    opls_used = []  # opls dihedrals that already were encountered
#    scaleDih = 0.0  # scale dihedrals with dummies

    for r in rlist:
        dih = rdic[r.resname][3]
        imp = rdic[r.resname][2]
        for d in imp+dih:
            al = []
            for name in d[:4]:
                if name.startswith('+'):
                    next = r.chain.fetch_residue(idx=r.id+1)
#                    next = r.chain.residues[idx+1]
                    atom = next.fetch(name[1:])[0]
                    al.append(atom)
                elif name.startswith('-'):
                    prev = r.chain.fetch_residue(idx=r.id-1)
#                    prev = r.chain.residues[idx-1]
                    atom = prev.fetch(name[1:])[0]
                    al.append(atom)
                else:
                    atom = r.fetch(name)[0]
                    al.append(atom)

            for dx in topol.dihedrals:
                func = dx[4]
                if (dx[0].id == al[0].id and
                   dx[1].id == al[1].id and
                   dx[2].id == al[2].id and
                   dx[3].id == al[3].id) or \
                   (dx[0].id == al[3].id and
                   dx[1].id == al[2].id and
                   dx[2].id == al[1].id and
                   dx[3].id == al[0].id):
                    # the following checks are needed for amber99sb*-ildn
                    # do not overwrite proper (type9) with improper (type4)
                    if 'default-star' in d[4] and dx[4] == 9:
                        continue
                    # is the dihedral already found for ILDN
                    encountered = 0
                    foobar = [dx[0].id, dx[1].id, dx[2].id, dx[3].id, d[4]]
                    encountered = _is_ildn_dih_encountered(ildn_used, foobar, encountered)
                    foobar = [dx[0].id, dx[1].id, dx[2].id, dx[3].id, d[5]]
                    encountered = _is_ildn_dih_encountered(ildn_used, foobar, encountered)
                    if encountered == 1:
                        continue
                    if 'tors' in d[4]:
                        if 'tors' not in dx[5]:
                            continue

                    # check for opls
                    encountered = 0
                    foobar = [dx[0].id, dx[1].id, dx[2].id, dx[3].id, d[4]]
                    encountered = _is_ildn_dih_encountered(opls_used, foobar, encountered)
                    foobar = [dx[0].id, dx[1].id, dx[2].id, dx[3].id, d[5]]
                    encountered = _is_ildn_dih_encountered(opls_used, foobar, encountered)
                    if encountered == 1:
                        continue
                    if 'dih_' in d[4]:
                        if 'dih_' not in dx[5]:
                            continue

                    A, B = _check_case(al[:4])
                    paramA = topol.BondedParams.get_dihedral_param(al[0].type,
                                                                   al[1].type,
                                                                   al[2].type,
                                                                   al[3].type,
                                                                   func)
                    paramB = topol.BondedParams.get_dihedral_param(al[0].typeB,
                                                                   al[1].typeB,
                                                                   al[2].typeB,
                                                                   al[3].typeB,
                                                                   func)

                    astate = []
                    bstate = []
                    backup_dx = dx[:]
                    backup_dx2 = dx[:]

                    if d[4] == 'default-A':  # amber99sb
                        if 'un' in d[5]:
                            backup_dx2.append('undefB')
                        dih_predef_default.append(backup_dx2)
                        astate = paramA
                    elif d[4] == 'default-B':  # amber99sb
                        if 'un' in d[5]:
                            backup_dx2.append('undefB')
                        dih_predef_default.append(backup_dx2)
                        astate = paramB
                    elif d[4] == 'default-star':  # amber99sb*
                        foo = [4, 105.4, 0.75, 1]
                        astate.append(foo)
                    elif d[4].startswith('torsion_'):  # amber99sb-ildn
                        if 'un' in d[5]:
                            backup_dx2.append('undefB_ildn')
                        dih_predef_default.append(backup_dx2)
                        foo = explicit_def[d[4]]
                        astate.append(foo)
                        foobar = [dx[0].id, dx[1].id, dx[2].id, dx[3].id, d[4]]
                        ildn_used.append(foobar)
                        func = 9
                    elif d[4].startswith('dih_'):  # opls proper
                        if 'un' in d[5]:
                            backup_dx2.append('undefB')
                        dih_predef_default.append(backup_dx2)
                        foo = explicit_def[d[4]]
                        astate.append(foo)
                        foobar = [dx[0].id, dx[1].id, dx[2].id, dx[3].id, d[4]]
                        opls_used.append(foobar)
                        func = 3
                    elif d[4].startswith('improper_'):  # opls improper
                        if 'un' in d[5]:
                            backup_dx2.append('undefB')
                        dih_predef_default.append(backup_dx2)
                        foo = explicit_def[d[4]]
                        astate.append(foo)
                        foobar = [dx[0].id, dx[1].id, dx[2].id, dx[3].id, d[4]]
                        opls_used.append(foobar)
                        func = 1
                    elif 'un' in d[4]:  # amber99sb-ildn and opls and others
                        if d[5].startswith('torsion_'):
                            backup_dx2.append('undefA_ildn')
                        else:
                            backup_dx2.append('undefA')
                        dih_predef_default.append(backup_dx2)
                        astate = ''
                    else:
                        astate = d[4]

                    if d[5] == 'default-A':  # amber99sb
                        bstate = paramA
                    elif d[5] == 'default-B':  # amber99sb
                        bstate = paramB
                    elif d[5] == 'default-star':  # amber99sb*
                        foo = [4, 105.4, 0.75, 1]
                        bstate.append(foo)
                    elif d[5].startswith('torsion_'):  # amber99sb-ildn
                        foo = explicit_def[d[5]]
                        bstate.append(foo)
                        foobar = [dx[0].id, dx[1].id, dx[2].id, dx[3].id, d[5]]
                        ildn_used.append(foobar)
                        func = 9
                    elif d[5].startswith('dih_'):  # opls proper
                        foo = explicit_def[d[5]]
                        bstate.append(foo)
                        foobar = [dx[0].id, dx[1].id, dx[2].id, dx[3].id, d[5]]
                        opls_used.append(foobar)
                        func = 3
                    elif d[5].startswith('improper_'):  # opls improper
                        foo = explicit_def[d[5]]
                        bstate.append(foo)
                        foobar = [dx[0].id, dx[1].id, dx[2].id, dx[3].id, d[5]]
                        ildn_used.append(foobar)
                        func = 1
                    elif 'un' in d[5]:  # amber99sb-ildn and opls and others
                        bstate = ''
                    else:
                        bstate = d[5]

                    # A state
                    counter = 0
                    for foo in astate:
                        if counter == 0:
                            dx[4] = func
                            if 'D' in A:
                               foo =  _scale_dih( foo, scaleDih ) 
                            dx[5] = foo
                            if(func == 3):
                                bar = [foo[0], 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
                            elif(func == 2):
                                bar = [foo[0], foo[1], 0.0]
                            else:
                                bar = [foo[0], foo[1], 0.0, foo[-1]]
                            dx.append(bar)
                        else:
                            alus = backup_dx[:]
                            if 'D' in A:
                               foo =  _scale_dih( foo, scaleDih )
                            alus[5] = foo
                            if func == 3:
                                bar = [foo[0], 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
                            elif(func == 2):
                                bar = [foo[0], foo[1], 0.0]
                            else:
                                bar = [foo[0], foo[1], 0.0, foo[-1]]
                            alus.append(bar)
                            dih9.append(alus)
                        counter = 1

                    # B state
                    for foo in bstate:
                        if counter == 0:
                            dx[4] = func
                            if(func == 3):
                                bar = [foo[0], 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
                            elif(func == 2):
                                bar = [foo[0], foo[1], 0.0]
                            else:
                                bar = [foo[0], foo[1], 0.0, foo[-1]]
                            dx[5] = bar
                            if 'D' in B:
                               foo =  _scale_dih( foo, scaleDih )
                            dx.append(foo)
                        else:
                            alus = backup_dx[:]
                            if 'D' in B:
                               foo =  _scale_dih( foo, scaleDih )
                            if(func == 3):
                                bar = [foo[0], 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
                            elif(func == 2):
                                bar = [foo[0], foo[1], 0.0]
                            else:
                                bar = [foo[0], foo[1], 0.0, foo[-1]]
                            alus[5] = bar
                            alus.append(foo)
                            dih9.append(alus)
                        counter = 1

    topol.dihedrals.extend(dih9)


def _get_hybrid_residue(residue_name, mtp_file='ffamber99sb.mtp',
                        version='new', verbose=False):
    if verbose is True:
        print('log_> Scanning database for %s ' % residue_name)
    resi, bonds, imps, diheds, rotdic = read_mtp_entry(residue_name,
                                                       filename=mtp_file,
                                                       version=version)
    if len(resi.atoms) == 0:
        raise mtpError("Hybrid residue %s not found in %s"
                       % (residue_name, mtp_file))
    return resi, bonds, imps, diheds, rotdic


def _get_hybrid_residues(m, ff, version='new', verbose=False):
    rdic = {}
    rlist = []
    for res in m.residues:
        if res.is_hybrid():
            rlist.append(res)
            # get topol params for hybrid
            mtp_file = get_mtp_file(res, ff)
            mtp = _get_hybrid_residue(residue_name=res.resname,
                                      mtp_file=mtp_file,
                                      version=version,
                                      verbose=verbose)
            rdic[res.resname] = mtp
            hybrid_res = mtp[0]
            atom_names = list(map(lambda a: a.name, hybrid_res.atoms))
            atoms = res.fetchm(atom_names)
            for i, atom in enumerate(hybrid_res.atoms):
                atoms[i].atomtypeB = atom.atomtypeB
                atoms[i].qB = atom.qB
                atoms[i].mB = atom.mB
    return rlist, rdic


def _add_extra_DNA_RNA_impropers(topol, rlist, func_type, stateA, stateB):
    extra_impropers = []
    for r in rlist:
        if r.resname in ['DAT', 'DAC', 'DGC', 'DGT', 'RAU', 'RAC', 'RGC', 'RGU']:
            print('Adding extra improper dihedrals for residue %d-%s'
                  % (r.id, r.resname))
            alist = r.fetchm(['C1\'', 'N9', 'C8', 'DC2'])
            extra_impropers.append(alist)
            alist = r.fetchm(['C1\'', 'N9', 'C4', 'DC6'])
            extra_impropers.append(alist)
        elif r.resname in ['DTA', 'DTG', 'DCG', 'DCA', 'RUA', 'RUG', 'RCG', 'RCA']:
            print('Adding extra improper dihedrals for residue %d-%s'
                  % (r.id, r.resname))
            alist = r.fetchm(['C1\'', 'N1', 'C6', 'DC4'])
            extra_impropers.append(alist)
            alist = r.fetchm(['C1\'', 'N1', 'C2', 'DC8'])
            extra_impropers.append(alist)
    for imp in extra_impropers:
        imp.extend([func_type, [func_type]+stateA, [func_type]+stateB])
    topol.dihedrals += extra_impropers


# =======
# Classes
# =======
class AbsRestraints:
    '''Identifies a set of restraints as defined by Boresch between the
    protein/host and ligand/guest molecule. [5]_ ::

                (Pro)  p3 -- p2          l2   (Lig)
                              \         /  \\
                               p1 --- l1    l3

    The process of choosing atoms involved in the restraints is
    stochastic. Thus, every time you call this function, you might get a
    different set of restraints. If you would like to have a deterministic
    behaviour provide a ``seed``.

    If you have a structure of the protein-ligand complex and want
    to choose restraints based on this, you can do so by providing a
    Model of the ``complex`` as well as the residue name of the ligand
    (``ligname``).

    If you know which atoms to use for the restraints, you can provide their
    atom indices as they are in the input Model objects with the arguments
    ``pro_ids`` and ``lig_ids``. The expected order of atoms is ``l1, l2, l3`` as
    in the scheme above for the ligand, and ``p1, p2, p3`` for the protein.

    Parameters
    ----------
    protein : Model, optional
        model of the protein/host molecule.
    ligand : Model, optional
        model of the ligand/guest molecule.
    complex: Model, optional
        model of the protein-ligand or host-guest complex.
    ligname : str, optional
        name of ligand/host residue. This is required if a complex is provided.
    pro_ids : list, optional
        list with three indeces.
    lig_ids : list, optional
        list with three indeces.
    kbond : float, optional
        force constant to use for the distance restraint.
        Default is 4184 kJ/(mol * nm^2).
    kangle : float, optional
        force constant to use for the angle restraints.
        Default is 41.84 kJ/(mol * rad^2).
    kdihedral : float, optional
        force constant to use for the dihedral restraints.
        Default is 41.84 kJ/(mol * rad^2).
    T : float, optional
        temperature in Kelvin. Default is 298.15 K.
    seed : bool, optional
        random seed.

    Attributes
    ----------
    lig_atoms : list
        list of three ligand/guest Atoms selected.
    pro_atoms : list
        list of three protein/host Atoms selected.
    dist : float
        distance ``l1-p1`` in nm.
    angle1 : float
        angle ``l2-l1-p1`` in degrees.
    angle2 : float
        angle ``l1-p1-p2`` in degrees.
    dihedral1 : float
        dihedral ``l3-l2-l1-p1`` in degrees.
    dihedral2 : float
        dihedral ``l2-l1-p1-p2`` in degrees.
    dihedral3 : float
        dihedral ``l1-p1-p2-p3`` in degrees.
    dg : float
        free energy of restraining the ligand while decoupled. This is calculate
        with equation 32 in Boresch et al. [5]_

    '''
    def __init__(self, protein=None, ligand=None, complex=None, ligname=None,
                 pro_ids=None, lig_ids=None,
                 kbond=4184, kangle=41.84, kdihedral=41.84,
                 T=298.15, seed=None):

        # check which scenario we have: either protein/ligand models are
        # provided separately, or a complex is provided along with the
        # residue name for the ligand.
        self.complex = deepcopy(complex)
        if self.complex is not None:
            if protein is not None or ligand is not None:
                raise ValueError('both arguments "complex" and "protein" or '
                                 '"ligand" have been provided. You should provide '
                                 'either separate protein and ligand Models '
                                 'or a single Model of the complex, but not '
                                 'both')
            if ligname is None:
                raise ValueError('when providing a protein-ligand complex '
                                 'you also need to provide the residue name '
                                 'of the ligand')
            self.ligname = ligname
            self.protein, self.ligand = self._split_complex_model()
        if self.complex is None:
            self.protein = deepcopy(protein)
            self.ligand = deepcopy(ligand)

        # other input parameters
        self.pro_ids = pro_ids
        self.lig_ids = lig_ids
        self.kbond = float(kbond)
        self.kangle = float(kangle)
        self.kdihedral = float(kdihedral)
        self.T = float(T)
        self.seed = seed
        # restraints data
        self.lig_atoms = []
        self.pro_atoms = []
        self.dist = None
        self.angle1 = None
        self.angle2 = None
        self.dihedral1 = None
        self.dihedral2 = None
        self.dihedral3 = None

        # check protein and ligand model are using the same units
        assert self.ligand.unity == self.protein.unity
        if self.ligand.unity == 'A':
            self.f = 10.0
        elif self.ligand.unity == 'nm':
            self.f = 1.0

        # pick atoms and define restraints
        self.select_restraints()

    def select_restraints(self):
        '''Automated restraints selection. For the ligand/guest, three heavy
        atoms close to its center of mass are selected. For the protein,
        the backbone N-CA-C atoms closest to the selected ligand atoms are
        picked. If the protein/host is another small molecule, then three
        atoms close to the ligand atoms are selected.
        '''
        # ===================
        # Choose Ligand Atoms
        # ===================

        # pick 3 atoms near the centre of mass
        if self.lig_ids is None:
            # find centre of mass
            com = Atom()
            com.x = self.ligand.com(vector_only=True)
            # calc distances from com, but exclude H atoms from selection
            distances = [(a, a-com) for a in self.ligand.atoms if a.symbol != 'H']
            # first atom picked is the one closest to centre of mass
            l1 = min(distances, key=lambda x: x[1])[0]
            # make a list of atoms within 3A of atom 1, excluding hydrogens
            atoms = [a for a in self.ligand.atoms if l1.id != a.id and l1-a < 0.3*self.f and a.symbol != 'H']
            # then pick two of them at random
            if self.seed is not None:
                np.random.seed(self.seed)
            l2, l3 = np.random.choice(atoms, size=2)
        else:
            if len(self.lig_ids) != 3:
                raise ValueError()
            l1 = self.ligand.fetch_atoms(key=self.lig_ids[0], how='byid')[0]
            l2 = self.ligand.fetch_atoms(key=self.lig_ids[1], how='byid')[0]
            l3 = self.ligand.fetch_atoms(key=self.lig_ids[2], how='byid')[0]

        self.lig_atoms = [l1, l2, l3]

        # ====================
        # Choose Protein Atoms
        # ====================
        if self.pro_ids is None:
            # check there are atoms close to the ligand (within 1 nm)
            atoms = [a for a in self.protein.atoms if a.symbol != 'H' and l1-a < 1.0*self.f]
            assert len(atoms) > 2
            # if host is a protein, choose backbone atoms
            # -------------------------------------------
            if self.protein.moltype == 'protein':
                N_atoms = [(a, l1-a) for a in atoms if a.name == 'N']
                # get closest N atom
                p1 = min(N_atoms, key=lambda x: x[1])[0]
                # get CA and C of the same residue
                mol = p1.molecule
                p2 = mol.fetch_atoms(key='CA', how='byname')[0]
                p3 = mol.fetch_atoms(key='C', how='byname')[0]

            # if host is not a protein, choose closest atom, plus 2 other random
            # ------------------------------------------------------------------
            else:
                atoms = [(a, l1-a) for a in self.protein.atoms if a.symbol != 'H']
                # first atom is the one closest to l1
                p1 = min(atoms, key=lambda x: x[1])[0]
                # make a list of atoms within 3A of atom 1, excluding hydrogens
                atoms = [a for a in self.protein.atoms if p1.id != a.id and p1-a < 0.3*self.f and a.symbol != 'H']
                # then pick two of them at random
                if self.seed is not None:
                    np.random.seed(self.seed)
                p2, p3 = np.random.choice(atoms, size=2)
        else:
            if len(self.pro_ids) != 3:
                raise ValueError()
            p1 = self.protein.fetch_atoms(key=self.pro_ids[0], how='byid')[0]
            p2 = self.protein.fetch_atoms(key=self.pro_ids[1], how='byid')[0]
            p3 = self.protein.fetch_atoms(key=self.pro_ids[2], how='byid')[0]

        self.pro_atoms = [p1, p2, p3]

        # --------------------------
        # Calculate distances/angles
        # --------------------------
        self.dist = (l1-p1)/self.f  # make sure it is in nm
        self.angle1 = l1.angle(l2, p1, degree=True)
        self.angle2 = p1.angle(l1, p2, degree=True)
        self.dihedral1 = l3.dihedral(l2, l1, p1, degree=True)
        self.dihedral2 = l2.dihedral(l1, p1, p2, degree=True)
        self.dihedral3 = l1.dihedral(p1, p2, p3, degree=True)

        # calc restraints dg
        self.calc_dg()

    def calc_dg(self, release=False):
        '''Calculates the free energy contribution of the
        restraints given the distance, angles, and dihedrals and their
        force constants as defined in the instance.

        Parameters
        ----------
        release : bool, optional
            whether to calculate the free energy of adding (False) or releasing
            (True) the restraints.
        '''
        V0 = 1.66  # standard volume in nm^3
        R = 8.314472*0.001  # Gas constant (kJ/mol/K)
        RT = R * self.T
        prefactor = (8.0*np.power(np.pi, 2.0) * V0 / (np.power(self.dist, 2.0) *
                     np.sin(self.angle1*np.pi/180.0) * np.sin(self.angle2*np.pi/180.0)))
        fconstants = (np.sqrt(self.kbond*self.kangle*self.kangle*self.kdihedral*self.kdihedral*self.kdihedral) /
                      np.power(2.0*np.pi*RT, 3.0))

        if release is True:
            self.dg = -RT * np.log(prefactor * fconstants)
        elif release is False:
            self.dg = RT * np.log(prefactor * fconstants)

    def make_ii(self, switch_on=True, ligand_first=True):
        '''Returns the data needed to add an intermolecular interactions
        section in a topology file to be used as restraints.

        Parameters
        ----------
        switch_on : bool, optional
            whether you want to switch the restraints on (restraints present in
            state B) or off (restraints present in state A). Default is
            True (switching on).
        ligand_first : bool, optional
            whether in the input gro/pdb file of the protein-ligand complex
            you are preparing the ligand atoms will come before (True) or after
            (False) the protein atoms. This is needed in order to adjust the
            atoms indices correctly. If you provided a Model of the complex
            (self.is_complex is True), this argument is disregarded as indices
            as provided will be used.

        Returns
        -------
        ii : dict
            dictionary with "bonds", "angles", and "dihedrals" keys with
            intermolecular_interactions information that can
            be passed to a Topology object.
        '''

        ii = {}
        ii['bonds'] = []
        ii['angles'] = []
        ii['dihedrals'] = []

        # ----------------
        # Sort out indices
        # ----------------
        l1 = self.lig_atoms[0].id
        l2 = self.lig_atoms[1].id
        l3 = self.lig_atoms[2].id
        p1 = self.pro_atoms[0].id
        p2 = self.pro_atoms[1].id
        p3 = self.pro_atoms[2].id
        if self.complex is None:
            if ligand_first is True:
                nshift = len(self.ligand.atoms)
                p1 += nshift
                p2 += nshift
                p3 += nshift
            elif ligand_first is False:
                nshift = len(self.protein.atoms)
                l1 += nshift
                l2 += nshift
                l3 += nshift

        # -----
        # Bonds
        # -----
        if switch_on is True:
            kA = 0.0
            kB = self.kbond
        elif switch_on is False:
            kA = self.kbond
            kB = 0.0

        ii['bonds'].append([l1, p1, 6, [self.dist, kA, self.dist, kB]])

        # ------
        # Angles
        # ------
        if switch_on is True:
            kA = 0.0
            kB = self.kangle
        else:
            kA = self.kangle
            kB = 0.0

        ii['angles'].append([l2, l1, p1, 1, [self.angle1, kA, self.angle1, kB]])
        ii['angles'].append([l1, p1, p2, 1, [self.angle2, kA, self.angle2, kB]])

        # ---------
        # Dihedrals
        # ---------
        if switch_on is True:
            kA = 0.0
            kB = self.kdihedral
        else:
            kA = self.kdihedral
            kB = 0.0

        ii['dihedrals'].append([l3, l2, l1, p1, 2, [self.dihedral1, kA, self.dihedral1, kB]])
        ii['dihedrals'].append([l2, l1, p1, p2, 2, [self.dihedral2, kA, self.dihedral2, kB]])
        ii['dihedrals'].append([l1, p1, p2, p3, 2, [self.dihedral3, kA, self.dihedral3, kB]])

        return ii

    def write_summary(self, fname='restraints.info'):
        '''Write restraints information, including free energy of restraining,
        to file.

        Parameters
        ----------
        fname : str, optional
            name of output file. Default is "restraints.info".
        '''

        summary = """================== Restraints ==================
atoms (lig)   = {0:>6} {1:>6} {2:>6}
atoms (pro)   = {3:>6} {4:>6} {5:>6}
""".format(self.lig_atoms[0].id, self.lig_atoms[1].id, self.lig_atoms[2].id,
           self.pro_atoms[0].id, self.pro_atoms[1].id, self.pro_atoms[2].id)

        summary += """------------------------------------------------
T             = {T:>14.2f} K
kbond         = {kbond:>14.2f} kJ/(mol * nm^2)
kangles       = {kangle:>14.2f} kJ/(mol * rad^2)
kdihedrals    = {kdihedral:>14.2f} kJ/(mol * rad^2)
r0            = {dist:>14.6f} nm
thA           = {angle1:>14.6f} deg
thB           = {angle2:>14.6f} deg
phiA          = {dihedral1:>14.6f} deg
phiB          = {dihedral2:>14.6f} deg
phiC          = {dihedral3:>14.6f} deg
================================================
dG Restraints = {dg:>14.2f} kJ/mol""".format(**self.__dict__)

        summary += "\n              = {0:>14.2f} kcal/mol\n".format(self.dg/4.184)

        with open(fname, 'w') as f:
            f.write(summary)

    def _split_complex_model(self):
        complex = deepcopy(self.complex)
        # find ligand
        ligands = complex.fetch_residues(key=self.ligname)
        # do a couple of checks
        if len(ligands) == 0:
            raise ValueError('no ligand with residue name %s found in Model '
                             'from file %s' % (self.ligname, self.complex.filename))
        elif len(ligands) > 1:
            raise ValueError('multiple ligands with residue name %s found in Model '
                             'from file %s' % (self.ligname, self.complex.filename))

        # make new Model for ligand
        ligand_mol = ligands[0]
        ligand = Model()
        for a in ligand_mol.atoms:
            ligand.atoms.append(a)
        ligand.unity = ligand_mol.atoms[0].unity

        # make protein by removing ligand residue
        complex.remove_residue(ligands[0], renumber_atoms=False, renumber_residues=False)

        # assign moltypes
        ligand.assign_moltype()
        complex.assign_moltype()

        return complex, ligand


# ==============
# Data/Libraries
# ==============
_dna_names = {
    'DA5_DT5': 'D5K',
    'DA5_DC5': 'D5L',
    'DA5_DG5': 'D5M',
    'DT5_DA5': 'D5N',
    'DT5_DC5': 'D5O',
    'DT5_DG5': 'D5P',
    'DC5_DA5': 'D5R',
    'DC5_DT5': 'D5S',
    'DC5_DG5': 'D5T',
    'DG5_DA5': 'D5X',
    'DG5_DT5': 'D5Y',
    'DG5_DC5': 'D5Z',
    'DA3_DT3': 'D3K',
    'DA3_DC3': 'D3L',
    'DA3_DG3': 'D3M',
    'DT3_DA3': 'D3N',
    'DT3_DC3': 'D3O',
    'DT3_DG3': 'D3P',
    'DC3_DA3': 'D3R',
    'DC3_DT3': 'D3S',
    'DC3_DG3': 'D3T',
    'DG3_DA3': 'D3X',
    'DG3_DT3': 'D3Y',
    'DG3_DC3': 'D3Z',
    # False names to avoid an error
    'DG3_DG3': 'FOO',
    'DC3_DC3': 'FOO',
    'DA3_DA3': 'FOO',
    'DT3_DT3': 'FOO',
    'DG5_DG5': 'FOO',
    'DC5_DC5': 'FOO',
    'DA5_DA5': 'FOO',
    'DT5_DT5': 'FOO',
    }

_rna_names = {
    'RA5_RU5': 'R5K',
    'RA5_RC5': 'R5L',
    'RA5_RG5': 'R5M',
    'RU5_RA5': 'R5N',
    'RU5_RC5': 'R5O',
    'RU5_RG5': 'R5P',
    'RC5_RA5': 'R5R',
    'RC5_RU5': 'R5S',
    'RC5_RG5': 'R5T',
    'RG5_RA5': 'R5X',
    'RG5_RU5': 'R5Y',
    'RG5_RC5': 'R5Z',
    'RA3_RU3': 'R3K',
    'RA3_RC3': 'R3L',
    'RA3_RG3': 'R3M',
    'RU3_RA3': 'R3N',
    'RU3_RC3': 'R3O',
    'RU3_RG3': 'R3P',
    'RC3_RA3': 'R3R',
    'RC3_RU3': 'R3S',
    'RC3_RG3': 'R3T',
    'RG3_RA3': 'R3X',
    'RG3_RU3': 'R3Y',
    'RG3_RC3': 'R3Z',
    # False names to avoid an error
    'RG3_RG3': 'FOO',
    'RC3_RC3': 'FOO',
    'RA3_RA3': 'FOO',
    'RU3_RU3': 'FOO',
    'RG5_RG5': 'FOO',
    'RC5_RC5': 'FOO',
    'RA5_RA5': 'FOO',
    'RU5_RU5': 'FOO',
    }



#!/usr/bin/env python

# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

"""Trajectory module
"""

from . import xdrfile


class Trajectory(xdrfile.XDRFile):
    """ changed Trajectory class to use new XDRFile class and keep old scripts working"""

    def __init__(self, filename, **kwargs):
        xdrfile.XDRFile.__init__(self, filename, **kwargs)

    def get_natoms(self):
        return self.natoms



import sys, os
from pmx import *
from pmx.options import *
from pmx.parser import *
from pmx.forcefield import *
from pmx.ndx import *
from string import digits
import copy as cp
import random
import numpy as np
import re

def split_itp_ffitp( itpfname, ffitpfname, itpoutfname ):
    itp = TopolBase( itpfname )
    fp = open(ffitpfname,'w')
    fp.write('[ atomtypes ]\n')
    for at in itp.atomtypes:
        fp.write('%8s %12.6f %12.6f %3s %12.6f %12.6f\n' % (at['name'],at['mass'],at['charge'],at['ptype'],at['sigma'],at['epsilon']))
    fp.close()
    itp.atomtypes = []
    itp.write(itpoutfname)

def structure_from_xyz( fname ):
    fp = open(fname,'r')
    lines = fp.readlines()
    fp.close()
    m = Model()
    counter = 0
    for l in lines[2:]:
        l = l.rstrip().lstrip().split()
        # add atom 
        a = Atom()
        a.name = l[0]
        a.resname = 'MOL'
        a.resnr = 1
        a.id = counter+1
        a.x[0] = float(l[1])
        a.x[1] = float(l[2])
        a.x[2] = float(l[3])
        m.atoms.append(a)
        counter+=1
    return(m)

def get_overall_charge_gaussian( fname ):
    fp = open(fname,'r')
    lines = fp.readlines()
    fp.close()
    for l in lines:
        if 'Charge' in l:
            l = l.rstrip()
            l = l.lstrip()
            foo = l.split()
            return(int(foo[2]))

def read_itp_top(fname):
    """ this function reads topology files from gaff """

    lines = open(fname).readlines()
    lines = kickOutComments(lines,';')
    itp = TopolBase(fname)#, ff = 'amber03')
    itp = Topology(fname,is_itp=True,assign_types=True,self_contained=True)#, ff = 'amber03')
#    atypes = read_atomtypes(lines, ff = 'amber03')
#    itp.atomtypes = atypes
    return itp

def write_ff(atypes, fname, ff='amber99sb'):
    fp = open(fname,'w')
    fp.write('[ atomtypes ]\n')
    for a in atypes:
        fp.write('%8s %12.6f %12.6f %3s %12.6f %12.6f\n' % (a['name'],a['mass'],a['charge'],a['ptype'],a['sigma'],a['epsilon']))
#    for atkey in atypes[0].keys():
#        at = atypes[atkey]

def set_ligname( ligname, m, itp ):
    # pdb
    for a in m.atoms:
        a.resname = ligname
    # itp
    for a in itp.atoms:
        a.resname = ligname
    itp.name = ligname

def gen_itp( fname_top, fname_gro, randnum ):
#    fname_itp = 'MOL.itp'
#    fname_ffitp = 'ffMOL.itp'
    itp = read_itp_top(fname_top)
#    itp.write(fname_itp)
#    write_ff(itp.atomtypes, fname_ffitp)
    return(itp)

def run_acpype_from_pdb( finp, ff, chargeMethod='bcc', charge=42 ):
    if charge==42: # guess the charge
        cmd = 'acpype -a {0} -o gmx -i {1} -c {2}'.format(ff,finp,chargeMethod)
    else:
        cmd = 'acpype -a {0} -o gmx -i {1} -c {2} -n {3}'.format(ff,finp,chargeMethod,charge)
    os.system(cmd)

def run_acpype( fname_prmtop, fname_inpcrd, charge, randnum, ff ):
    fname_top = 'MOL_GMX.top'
    fname_gro = 'MOL_GMX.gro'
    cmd = 'acpype -p '+fname_prmtop+' -x '+fname_inpcrd+' -o gmx -a '+ff+' -n '+str(charge)+' -c user -b MOL'
    os.system(cmd)
    return(fname_top,fname_gro)

def run_teleap( fname_mol2, fname_frcmod, randnum, ff ):
    fname_prmtop = 'prmtop_'+str(randnum)
    fname_inpcrd = 'inpcrd_'+str(randnum)
    # input file
    fp = open('leap.in','w')
#    fp.write("source %s/dat/leap/cmd/leaprc.%s\n" % (os.environ['AMBERHOME'],ff) )
    fp.write("loadamberparams %s/dat/leap/parm/%s.dat\n" % (os.environ['AMBERHOME'],ff) )
    fp.write("M = loadmol2 %s\n" % fname_mol2)
    fp.write("mods = loadamberparams %s\n" % fname_frcmod)
    fp.write("saveamberparm M %s %s\n" % (fname_prmtop,fname_inpcrd) )
    fp.write("quit\n")
    fp.close()
    # run teleap
    cmd = "teLeap -f leap.in"
    os.system(cmd)
    return(fname_prmtop,fname_inpcrd)

def generate_mol2( fname_ac_resp, halogens, halogens_nn, sigmaholes, randnum, ff, charge ):
    fname_mol2 = 'mol2_'+str(randnum)+'.mol2'
    # generate mol2
    # 1. attempt
    cmd = 'antechamber -i '+fname_ac_resp+' -fi ac -o '+fname_mol2+' -fo mol2 -at '+ff+' -nc '+str(charge)+' -j 4 -dr no >> /dev/null 2>&1'
    os.system(cmd)
    # 2. attempt (if the first attempt failed, try flag -j 5)
    if os.path.isfile(fname_mol2)==False:
        cmd = 'antechamber -i '+fname_ac_resp+' -fi ac -o '+fname_mol2+' -fo mol2 -at '+ff+' -nc '+str(charge)+' -j 5 -dr no >> /dev/null 2>&1'
        os.system(cmd)

    # add bonds
    fp = open(fname_mol2,'r')
    lines = fp.readlines()
    fp.close()
    mol2 = {}
    mol2['header'] = readSection(lines,'@<TRIPOS>MOLECULE','@')
    mol2['atoms'] = readSection(lines,'@<TRIPOS>ATOM','@')
    mol2['bonds'] = readSection(lines,'@<TRIPOS>BOND','@')
    mol2['substr'] = readSection(lines,'@<TRIPOS>SUBSTRUCTURE','@')
    fp = open(fname_mol2,'w')
    fp.write('@<TRIPOS>MOLECULE\n')
    for l in mol2['header']:
        fp.write(l)
    fp.write('@<TRIPOS>ATOM\n')
    for l in mol2['atoms']:
        foo = l.split()
        # check if sigmahole
        bSH = False
        for sh in sigmaholes:
            if sh.id==int(foo[0]):
                bSH = True
                break
        if bSH==True:
            fp.write("%7s%5s%15s%11s%11s%3s%10s%4s%15s\n" % (foo[0],foo[1],foo[2],foo[3],foo[4],'DU',foo[6],foo[7],foo[8]) )
        else: 
            fp.write(l)
    fp.write('@<TRIPOS>BOND\n')
    foo = ""
    for l in mol2['bonds']:
        fp.write(l)
        foo = l.split()
    bondnum = int(foo[0])
#    for h,hnn,sh in zip(halogens,halogens_nn,sigmaholes):
#        fp.write("%6s%6s%6s%2s\n" % (bondnum,h.id,sh.id,1) )
#        bondnum+=1
    fp.write('@<TRIPOS>SUBSTRUCTURE\n')
    for l in mol2['substr']:
        fp.write(l)
    fp.close()
    return(fname_mol2)

def read_frcmod( fname ):
    fp = open(fname,'r')
    lines = fp.readlines()
    fp.close()
    out = {}
    out['MASS'] = readSection(lines,'MASS','BOND')
    out['BOND'] = readSection(lines,'BOND','ANGLE')
    out['ANGLE'] = readSection(lines,'ANGLE','DIHE')
    out['DIHE'] = readSection(lines,'DIHE','IMPROPER')
    out['IMPROPER'] = readSection(lines,'IMPROPER','NONBON')
    out['NONBON'] = readSection(lines,'NONBON','*')
    return(out)

def gen_parmchk_dum( randnum ):
    fname = 'parmchk_'+str(randnum)
    fp = open(fname,'w')
    fp.write('PARM    DU      1               1        0.00   0       0\n')
    fp.close()
    return( fname )

def generate_frcmod( fname_ac_resp, halogens, halogens_nn, randnum, rule, ff, scaleD=1.0 ):
    fname_frcmod = 'frcmod_'+str(randnum)
    # from AmberTools20 need to generate an additional parmchk.dat file for dummies
    add_parmchk_fname = gen_parmchk_dum( randnum )
    # run frcmod
    cmd = 'parmchk2 -i {0} -f ac -o {1} -s {2} -atc {3}'.format(fname_ac_resp,fname_frcmod,ff,add_parmchk_fname)
    os.system(cmd)
    # correct frcmod
    frcmod = read_frcmod( fname_frcmod )
    fp = open(fname_frcmod,'w')
    fp.write('remark\n')
    # mass
    fp.write('MASS\n')
    for l in frcmod['MASS']:
        fp.write(l)
    # bonds
    fp.write('BOND\n')
    for l in frcmod['BOND']:
        if 'DU' in l:
            foo = l[0:5]#l.split()[0]
            l = foo+'       600.0 '+str(get_hal_sigma_distance( rule, foo, scaleD=scaleD ))+'\n'
        fp.write(l)
    # angles
    fp.write('\nANGLE\n')
    for l in frcmod['ANGLE']:
        if 'DU' in l:
            foo = l[0:8]#l.split()[0]
            l = foo+'           150.0     180.00\n'
        fp.write(l)
    # dihedrals
    fp.write('\nDIHE\n')
    for l in frcmod['DIHE']:
        if 'DU' in l:
            foo = l[0:11]#l.split()[0]
            l = foo+'     1    0.000         0.000           0.000\n'
        fp.write(l)
    # impropers
    fp.write('\nIMPROPER\n')
    for l in frcmod['IMPROPER']:
        fp.write(l)
    # nonbondeds
    fp.write('\nNONBON\n')
    for l in frcmod['NONBON']:
        if 'DU' in l:
            l = 'DU 1.0 0.0\n'
        fp.write(l)
    fp.close()
    return(fname_frcmod)

def antechamber_ac_resp( fname_ac, fname_resp, halogens, halogens_nn, sigmaholes, randnum, ff, charge ):
    fname_ac_resp = 'resp_'+str(randnum)+'.ac'
    # incorporate charges into .ac
    # 1. attempt
    cmd = 'antechamber -i '+fname_ac+' -fi ac -o '+fname_ac_resp+' -fo ac -c rc -cf '+fname_resp+' -j 4 -dr no -at '+ff+' -nc '+str(charge)+' > /dev/null 2>&1'
    os.system(cmd)
    # 2. attempt (if the first attempt failed, try flag -j 5)
    if os.path.isfile(fname_ac_resp)==False:
        cmd = 'antechamber -i '+fname_ac+' -fi ac -o '+fname_ac_resp+' -fo ac -c rc -cf '+fname_resp+' -j 5 -dr no -at '+ff+' -nc '+str(charge)+' > /dev/null 2>&1'
        os.system(cmd)

    # add bonds
#    add_bonds_ac( fname_ac_resp, halogens, sigmaholes )

    # in the ac file, sigmaholes have type LP, change it to DU (otherwise teleap will complain)
    ac_change_LP_to_DU( fname_ac_resp )

    return(fname_ac_resp)

def generate_respfiles( fname_ac, fname_esp, randnum ):
    fname_resp = 'respfile_'+str(randnum)
    # respgen
    cmd = 'respgen -i '+fname_ac+' -f resp1 -o respin1_'+str(randnum)+'.respin'
    os.system(cmd)
    cmd = 'respgen -i '+fname_ac+' -f resp2 -o respin2_'+str(randnum)+'.respin'
    os.system(cmd)
    # fit resp charges
    cmd = 'resp -O -i respin1_'+str(randnum)+'.respin -o respout1_'+str(randnum)+'.respout -e '+fname_esp+' -t qout_stage1'
    os.system(cmd)
    cmd = 'resp -O -i respin2_'+str(randnum)+'.respin -o respout2_'+str(randnum)+'.respout -e '+fname_esp+' -q qout_stage1 -t '+fname_resp
    os.system(cmd)
    return(fname_resp)

def generate_esp( fname_log, halogens, halogens_nn, m, randnum, rule, ff, scaleD=1.0, bESP=False ):
    fname_esp = 'espgen_'+str(randnum)+'.esp'

    # run espgen
    if bESP==False:
        cmd = 'espgen -i '+fname_log+' -o '+fname_esp
    else: # esp is provided as input
        cmd = 'cp {0} {1}'.format(fname_log,fname_esp)
    os.system(cmd)

    # insert sigmahole into esp
    fp = open(fname_esp,'r')
    lines = fp.readlines()
    fp.close()

    # extract coordinates from esp
    espcoord = extract_esp_coords( lines )

    # find halogen coords
    hcoords = []
    hnncoords = []
    for ah,ahnn in zip(halogens,halogens_nn):
        hcoords.append(espcoord[ah.id-1])
        hnncoords.append(espcoord[ahnn.id-1])

    # write esp file with inserted sigmahole
    fp = open(fname_esp,'w')
    bohr = 0.529177249
    i = 0
    bHalFinished = False
    for l in lines:
        lsplit = l.split()
        if i==0:
            foo = l[:5] # first 5 elements
            bar = l[5:]
            foo = foo.lstrip()
            foo = int(foo)+len(halogens)
            fp.write("%5d%s" % (foo,bar) )
            i+=1
            continue
        elif i>0 and len(lsplit)==4 and bHalFinished==False: # write halogens
            hnum = 0
            for ah,ahnn in zip(halogens,halogens_nn):
                ch = [ float(hcoords[hnum][0]),float(hcoords[hnum][1]),float(hcoords[hnum][2]) ]
                chnn = [ float(hnncoords[hnum][0]),float(hnncoords[hnum][1]),float(hnncoords[hnum][2]) ]
                vec = [ch[0]-chnn[0],ch[1]-chnn[1],ch[2]-chnn[2]]
                vecnorm = np.sqrt(vec[0]*vec[0]+vec[1]*vec[1]+vec[2]*vec[2])
                vec = vec/vecnorm
                halname = ah.name.lower()
                dist = get_hal_sigma_distance( rule, halname, scaleD=scaleD )
                newpos = [ch[0]+vec[0]*dist/bohr , ch[1]+vec[1]*dist/bohr, ch[2]+vec[2]*dist/bohr ]

                fp.write("%32.7E%16.7E%16.7E\n" % (newpos[0],newpos[1],newpos[2]) )
                hnum += 1
            bHalFinished = True
        fp.write(l)
        i+=1
    fp.close()
    return(fname_esp)

def extract_esp_coords( lines ):
    coord = []
    for l in lines[1:]:
        foo = l.split()
        if len(foo)>3:
            return(coord)
        coord.append(foo)

def get_halogen_position( ah, ahnn, rule, scaleD=1.0 ): 
    vec = [ah.x[0]-ahnn.x[0],ah.x[1]-ahnn.x[1],ah.x[2]-ahnn.x[2]]
    vecnorm = np.sqrt(vec[0]*vec[0]+vec[1]*vec[1]+vec[2]*vec[2])
    vec = vec/vecnorm
    halname = ah.name.lower()
    dist = get_hal_sigma_distance( rule, halname, scaleD=scaleD )
    newpos = [ah.x[0]+vec[0]*dist , ah.x[1]+vec[1]*dist, ah.x[2]+vec[2]*dist ]
    return(newpos)

def get_ac_atom_dict( lines ):
    out = {}
    for l in lines:
        if l.startswith('ATOM'):
            l = l.rstrip()
            foo = l.split()
            out[int(foo[1])] = foo[2]
    return(out)

def add_bonds_ac( fname_ac, halogens, sigmaholes):
    # read
    fp = open(fname_ac,'r')
    lines = fp.readlines()
    fp.close()
    # get bond number
    foo = lines[-1].split()
    bondnum = int(foo[1])
    # get atom name dictionary: dict[id] = at_name
    ac_atom_dict = get_ac_atom_dict( lines )
    # append new bonds
    fp = open(fname_ac,'a')
    i = 0
    for h,sh in zip(halogens,sigmaholes):
        bondnum+=1
        halname = ac_atom_dict[h.id]
        shname = ac_atom_dict[sh.id]
        fp.write("%4s%5d%5d%5d%5d%7s%5s\n" % ("BOND",bondnum,h.id,sh.id,1,halname,shname) )
        i+=1
    fp.close()

def ac_change_LP_to_DU( fname_ac ):
    regexp = re.compile('LP')
    # read
    fp = open(fname_ac,'r')
    lines = fp.readlines()
    fp.close()
    # write
    fp = open(fname_ac,'w')
    for l in lines:
        if l.startswith('ATOM'):
            if 'EP' in l:
                l = regexp.sub('DU',l)
        fp.write(l)
    fp.close()

def ac_change_DU_to_LP( fname_ac ):
    regexp = re.compile('DU')
    # read
    fp = open(fname_ac,'r')
    lines = fp.readlines()
    fp.close()
    # write
    fp = open(fname_ac,'w')
    for l in lines:
        if l.startswith('ATOM'):
            if 'EP' in l:
                l = regexp.sub('LP',l)
        fp.write(l)
    fp.close()

def antechamber_ac( fname_pdb, m, halogens, sigmaholes, charge, randnum, ff ):
    fname_ac = 'sigmahole_'+str(randnum)+'.ac'
    resname = 'MOL' #m.residues[0].resname
    # 1. attempt
    cmd = 'antechamber -i '+fname_pdb+' -fi pdb -o '+fname_ac+' -fo ac -j 4 -rn '+resname+' -nc '+str(charge)+' -dr no'+' -at '+ff+' > /dev/null 2>&1'
    os.system(cmd)
    # 2. attempt (if the first attempt failed, try flag -j 5)
    if os.path.isfile(fname_ac)==False:
        cmd = 'antechamber -i '+fname_pdb+' -fi pdb -o '+fname_ac+' -fo ac -j 5 -rn '+resname+' -nc '+str(charge)+' -dr no'+' -at '+ff+' > /dev/null 2>&1'
        os.system(cmd)

    # add bonds
    add_bonds_ac( fname_ac, halogens, sigmaholes )

#    ac_change_DU_to_LP( fname_ac )

    return(fname_ac)

def antechamber_pdb( fname, charge, randnum, ff ):
    fname_pdb = 'lig_gaussian_'+str(randnum)+'.pdb'
    resname = 'MOL' #m.residues[0].resname
    cmd = 'antechamber -i '+fname+' -fi gout -o '+fname_pdb+' -fo pdb -j 4 -rn '+resname+' -nc '+str(charge)+' -dr no'+' -at '+ff+' > /dev/null 2>&1'
    os.system(cmd)
    return(fname_pdb)

def add_sigmahole_pdb( halogens, halogens_nn, m, randnum, rule, ff, scaleD=1.0 ):
    sigmahole_num = 1
    sigmaholes = []
    fname_pdb = 'str_'+str(randnum)+'.pdb'
    for ah,ahnn in zip(halogens,halogens_nn):
        newpos = get_halogen_position( ah, ahnn, rule, scaleD=scaleD ) 

        newatom = cp.deepcopy(ah)
        newatom.name = 'EP'+str(sigmahole_num)
        newatom.x = newpos
        newatom.id = len(m.atoms)+1
        m.atoms.append(newatom)
        sigmaholes.append(newatom)

        fname_pdb = 'sigmahole_'+str(randnum)+'.pdb'
        sigmahole_num += 1

    m.write(fname_pdb)
    return(fname_pdb,sigmaholes)

def add_ffitp_sigmahole( ffitp, hid, hnnid, num, ff ):
    bFound = False
    for a in ffitp.atomtypes:
        if 'DU' in a['name']:
            bFound = True

    if bFound==False:
        atomtype = dict()
        atomtype['name'] = str('DU')
        atomtype['bond_type'] = str('DU')
        atomtype['mass'] = float(0.0)
        atomtype['charge'] = float(0.0)
        atomtype['ptype'] = str('A')
        atomtype['sigma'] = float(0.0)
        atomtype['epsilon'] = float(0.0)
        ffitp.atomtypes.append(atomtype)

#    if 'DU' not in ffitp.atomtypes.keys():
#        if 'opls' in ff:
#            ffitp.atomtypes['DU'] = [0.0,0.0,'A',0.0,0.0]
#        if 'cgenff' in ff:
#            ffitp.atomtypes['DU'] = [0.0,0.0,'A',0.0,0.0]
            # the other parameters are already incorporated into CGenFF: nbfix and LJ changes for halogens
#        if 'gaff' in ff:
#            ffitp.atomtypes['DU'] = [0.0,0.0,'A',0.0,0.0]

def add_itp_sigmahole( itp, hid, hnnid, num, rule, ff, scaleD=1.0 ):
    ah = itp.atoms[hid-1] # halogen
    ahnn = itp.atoms[hnnid-1] # atom to which halogen is bound (halogen nearest neighbor)
    halname = ah.name.lower()
    newatom = cp.deepcopy(ah)
    newatom.name = 'EP'+str(num)
    newatom.cgnr = len(itp.atoms)+1

    ########## atom ###########
    if 'cl' in halname:        
        if 'jorgensen' in rule:
            newatom.q = 0.075
            ah.q -= newatom.q
        elif 'gutierrez' in rule:
            newatom.q = 0.05
            ah.q -= 0.08
            ahnn.q += 0.03
        if 'ibrahim' in rule:
            newatom.q = 0.035
            ah.q -= newatom.q
    elif 'br' in halname:
        if 'jorgensen' in rule:
            newatom.q = 0.1
            ah.q -= newatom.q
        elif 'gutierrez' in rule:
            newatom.q = 0.05
            ah.q -= 0.08
            ahnn.q += 0.03
        if 'ibrahim' in rule:
            newatom.q = 0.04
            ah.q -= newatom.q
    elif 'i' in halname:
        if 'jorgensen' in rule:
            newatom.q = 0.11
            ah.q -= newatom.q
        elif 'gutierrez' in rule:
            newatom.q = 0.05
            ah.q -= 0.09
            ahnn.q += 0.04
        if 'ibrahim' in rule:
            newatom.q = 0.055
            ah.q -= newatom.q
    newatom.m = 0.0

    newatom.id = len(itp.atoms)+1
    newatom.atomtype = 'DU'
    itp.atoms.append(newatom)

    ######### bond ##########
    d = get_hal_sigma_distance( rule, halname, scaleD=scaleD )/10.0 # in nm

#    newbond = [hid,newatom.id,1,d,0.0]
    newbond = [ah,newatom,1,[np.round(d,4),0.0]]
    itp.bonds.append(newbond)
    # also add a bond to the halogen nearest neighbor:
    # this is done to exclude 1-2 and 1-3 interactions for the virtual particle
    # in the same way as it is done for the halogen atom,
    # because the virtual particle is considered a part of halogen
    bondlength = get_bond_length( itp, hid, hnnid )
#    newbond2 = [hnnid,newatom.id,1,bondlength+d,0.0]
    newbond2 = [ahnn,newatom,1,[np.round(bondlength+d,4),0.0]]
    itp.bonds.append(newbond2)

    ######### pairs #########
#    gen_pairs( itp, hid, hnnid, newatom.id )
    gen_pairs_alt( itp, hid, newatom ) # generate pairs as if sigma hole is the same atom as halogen

    ######### virtual site ##
    hnnid2 = 1
    dvsite = get_vsite_param( halname, ff, itp, hid, hnnid, rule, scaleD=scaleD ) 
    vsite = [newatom,ahnn,ah,1,dvsite]
    itp.has_vsites2 = True
    itp.virtual_sites2.append(vsite)

def check_and_add_bond( itp, a1, a2, d ):
    ids = [a1.id,a2.id]
    for b in itp.bonds:
        if (b[0].id in ids) and (b[1].id in ids):
            return(True)
    # bond not found
    newbond = [a1,a2,1,[np.round(d,4),999.9]] 
    itp.bonds.append(newbond)

def add_vsite_respfit( ff, itp, halogens, halogens_nn, sigmaholes, rule, scaleD=1.0 ):
    for ah,ahnn,sh in zip(halogens,halogens_nn,sigmaholes):
        #d_hal_sigma = (ah-sh)/10.0 # convert to nm
        d_hal_sigma = get_hal_sigma_distance( rule, ah.name.lower(), scaleD=scaleD )/10.0 # in nm
        # check if this bond is already in .itp, if not, add it
        check_and_add_bond( itp, ah, sh, d_hal_sigma )
#        d_hal_hnn = (ah-ahnn)/10.0 # convert to nm
        d_hal_hnn = get_bond_length( itp, ah.id, ahnn.id )
        dvsite = (d_hal_hnn+d_hal_sigma)/d_hal_hnn
        vsite = [sh,ahnn,ah,1,dvsite]
        itp.has_vsites2 = True
        itp.virtual_sites2.append(vsite)
#    return(itp)

def get_vsite_nums( itp ):
    dums = []
    for a in itp.atoms:
        if a.name.startswith('EP'):
            dums.append(a.id)
    return(dums)

def remove_angles_dihedrals( itp ):
    dums = get_vsite_nums( itp )
    
    # remove angles
    newangles = []
    for ang in itp.angles:
        a1 = ang[0].id
        a2 = ang[1].id
        a3 = ang[2].id
        if (a1 in dums) or (a2 in dums) or (a3 in dums):
            continue
        else:
            newangles.append(ang)
    itp.angles = newangles

    # remove dihedrals
    newdihedrals = []
    for dih in itp.dihedrals:
        a1 = dih[0].id
        a2 = dih[1].id
        a3 = dih[2].id
        a4 = dih[3].id
        if (a1 in dums) or (a2 in dums) or (a3 in dums) or (a4 in dums):
            continue
        else:
            newdihedrals.append(dih)
    itp.dihedrals = newdihedrals

def add_sh_hnn_bond_respfit( ff, itp, halogens, halogens_nn, sigmaholes ):
    for ah,ahnn,sh in zip(halogens,halogens_nn,sigmaholes):
#        d_hnn_sigma = (ahnn-sh)/10.0 # convert to nm
        # extract these distances from .itp, as the structure may not match equilibrium bond
        d_hal_hnn = get_bond_length( itp, ah.id, ahnn.id ) 
        d_hal_sh = get_bond_length( itp, ah.id, sh.id ) 
        d_hnn_sigma = d_hal_hnn + d_hal_sh
#        newbond = [ahnn.id,sh.id,1,d_hnn_sigma,999.99]
        newbond = [ahnn,sh,1,[d_hnn_sigma,999.99]]
        itp.bonds.append(newbond)
#    return(itp)

def add_pairs_respfit( itp, halogens, sigmaholes ):
    newpairs = cp.deepcopy(itp.pairs)
    pairstoremove = []
    pairstoadd = []
    for hal,sh in zip(halogens,sigmaholes):
        for p in newpairs:
            if p[0].id==sh.id or p[1].id==sh.id:
                pairstoremove.append(p)
                continue
            if p[0].id==hal.id:
                newpair = [sh,p[1],p[2]]
            elif p[1].id==hal.id:
                newpair = [p[0],sh,p[2]]
            else:
                continue
            pairstoadd.append(newpair)
    # remove
    for p in pairstoremove:
        newpairs.remove(p)
    # add
    for p in pairstoadd:
        newpairs.append(p)
    itp.pairs = newpairs

def get_bond_length( itp, hid, hnnid ):
    for b in itp.bonds:
        if b[0].id==hnnid and b[1].id==hid:
            return(b[3][0])
        elif b[0].id==hid and b[1].id==hnnid:
            return(b[3][0])
    return(0.0)

def gen_pairs_alt( itp, hid, newat ):
    pairat = []
    for p in itp.pairs:
        if p[0].id==hid:
            pairat.append(p[1])
        if p[1].id==hid:
            pairat.append(p[0])
    for p in pairat:
        newpair = [newat, p, 1]
        itp.pairs.append(newpair)

def gen_pairs( itp, hid, hnnid, newid ):
    pairid = []
    for b in itp.bonds:
        if b[0]==hnnid and b[1]!=hid:
            pairid.append(b[1])
        elif b[0]!=hid and b[1]==hnnid:
            pairid.append(b[0])
    for p in pairid:
        newpair = [newid, p, 1]
        itp.pairs.append(newpair)
        

def get_vsite_param( halname, ff, itp, hid, hnnid, rule, scaleD=1.0 ):
    vs = 0.0
    d_hal_sigma = get_hal_sigma_distance( rule, halname, scaleD=scaleD )/10.0 # converting to nm
    d_c_hal = 0.0
    for b in itp.bonds:
        if (b[0].id==hid and b[1].id==hnnid) or (b[0].id==hnnid and b[1].id==hid):
            d_c_hal = b[3][0]
    if d_c_hal==0.0:
        print("Something went wrong: no halogen binding atom was found")
        sys.exit(0)
    vs = (d_c_hal+d_hal_sigma)/d_c_hal 
    return(vs)

def get_hal_sigma_distance( rule, halname, scaleD=1.0 ):
    # all in Angstroms
    d = 0.0
    if 'cl' in halname:
        if 'jorgensen' in rule:
            d = 1.6
        elif 'gutierrez' in rule:
            d = 1.64
        if 'ibrahim' in rule:
            d = 1.945
    elif 'br' in halname:
        if 'jorgensen' in rule:
            d = 1.6
        elif 'gutierrez' in rule:
            d = 1.89
        if 'ibrahim' in rule:
            d = 2.02
    elif 'i' in halname:
        if 'jorgensen' in rule:
            d = 1.8
        elif 'gutierrez' in rule:
            d = 2.20
        if 'ibrahim' in rule:
            d = 2.15
    d = scaleD*d
    return(d)

def add_pdb_sigmahole( m, hid, hnnid, num, rule, ff, scaleD=1.0 ):
    ah = m.atoms[hid-1]
    ahnn = m.atoms[hnnid-1]
    vec = [ah.x[0]-ahnn.x[0],ah.x[1]-ahnn.x[1],ah.x[2]-ahnn.x[2]]
    vecnorm = np.sqrt(vec[0]*vec[0]+vec[1]*vec[1]+vec[2]*vec[2])
    vec = vec/vecnorm
    halname = ah.name.lower()

    dist = get_hal_sigma_distance( rule, halname, scaleD=scaleD )
    newpos = [ah.x[0]+vec[0]*dist , ah.x[1]+vec[1]*dist, ah.x[2]+vec[2]*dist ]

    newatom = cp.deepcopy(ah)
    newatom.name = 'EP'+str(num)
    newatom.x = newpos   
    newatom.id = len(m.atoms)+1
    m.atoms.append(newatom)

def gen_sigmaholes( itp, m, ffitp, halogens, halogens_nn, rule, ff='oplsaa', scaleD=1.0 ):
    for i in range(0,len(halogens)):
        hal = halogens[i]
        halnn = halogens_nn[i]
        hal_id = hal.id
        halnn_id = halnn.id
        sigmahole_num = i+1

        add_pdb_sigmahole( m, hal_id, halnn_id, sigmahole_num, rule, ff, scaleD=scaleD )
        add_itp_sigmahole( itp, hal_id, halnn_id, sigmahole_num, rule, ff, scaleD=scaleD )
        add_ffitp_sigmahole( ffitp, hal_id, halnn_id, sigmahole_num, ff )
    

def find_halogens( itp, m=False, bSH=True ):
    halogens = []
    halogens_nn = []
    if bSH==False:
        return(halogens,halogens_nn)

    if m==False:
        atoms = itp.atoms
    else:
        atoms = m.atoms

    for at in atoms:
        foo = at.name.lower()
        foo = foo.translate(str.maketrans('','',digits))
        if foo.startswith('cl') or foo.startswith('br') or foo.startswith('i'):
            halogens.append( at )

            if m==False:
                for b in itp.bonds:
                    a1 = b[0]
                    a2 = b[1]
                    if a1.id==at.id:
                        halogens_nn.append( itp.atoms[a2.id-1] )
                    elif a2.id==at.id:
                        halogens_nn.append( itp.atoms[a1.id-1] )
            else: # find neighbor in pdb based on distance
                mind = 9999.999
                nn = ""
                for a in atoms:
                    aname = a.name.lower()
                    aname = aname.translate(str.maketrans('','',digits))
                    if aname.startswith('h') or (aname==foo):
                        continue
                    d = (at.x[0]-a.x[0])**2 + (at.x[1]-a.x[1])**2 + (at.x[2]-a.x[2])**2
                    if d<mind:
                        mind = d
                        nn = a
                halogens_nn.append( nn )

    return( halogens,halogens_nn )

def write_ffitp(atypes, itp, fname, ff='opls'):
    fp = open(fname,'w')
    fp.write('[ atomtypes ]\n')
    ## output ffitp, if ffitp was read ##
    for atkey in atypes.keys():
        at = atypes[atkey]
        if 'opls' in ff:
            fp.write('%8s %12.6f %12.6f %3s %12.6f %12.6f\n' % (atkey,at[0], at[1], at[2], at[3], at[4]) )
        elif 'cgenff' in ff:
            fp.write('%8s %12.6f %12.6f %3s %12.6f %12.6f\n' % (atkey,at[0], at[1], at[2], at[3], at[4]) )
        elif 'gaff' in ff:
            fp.write('%8s %12.6f %12.6f %3s %12.6f %12.6f\n' % (atkey,at[0], at[1], at[2], at[3], at[4]) )
  #      else:
   #         print >>fp, '%8s %12.6f %12.6f %3s %12.6f %12.6f' % (at[0], at[1], at[2], at[3], at[4], at[5])
    ## also output sigmahole dummies, if present ##
#    for a in itp.atoms:
#        if 'DU' in a.atomtype:
#            if 'opls' in ff:
#                print >>fp, '%8s %12.6f %12.6f %3s %12.6f %12.6f' % (atkey,0.0, 0.0, 0.0, 0.0, 0.0)
#            elif 'cgenff' in ff:
#                print >>fp, '%8s %12.6f %12.6f %3s %12.6f %12.6f' % (atkey,0.0, 0.0, 0.0, 0.0, 0.0)
#            elif 'gaff' in ff:
#                print >>fp, '%8s %12.6f %12.6f %3s %12.6f %12.6f' % (atkey,0.0, 0.0, 0.0, 0.0, 0.0)

################################################################################33


def main(argv):

    version = "1.1"

	# define input/output files
    files= [
	   FileOption("-itp", "r/o",["itp"],"MOL.itp",""),
	   FileOption("-ffitp", "r/o",["itp"],"ffMOL.itp",""),
	   FileOption("-pdb", "r/o",["pdb"],"mol.pdb",""),
	   FileOption("-xyz", "r/o",["xyz"],"orca.xyz",""),
	   FileOption("-log", "r/o",["log"],"gaussian.log",""),
	   FileOption("-esp", "r/o",["esp"],"esp.esp",""),
	   FileOption("-oitp", "o",["itp"],"outMOL.itp",""),
	   FileOption("-offitp", "o",["itp"],"outffMOL.itp" ,""),
	   FileOption("-opdb", "o",["pdb"],"outpdb.pdb","" ),
	    ]

	# define options
    options=[
           Option( "-ff", "string", "opls", "force-field: opls, cgenff, gaff"),
           Option( "-rule", "string", "default", "ibrahim, jorgensen, gutierrez"),
           Option( "-ligname", "string", "", "set ligand name (if none given, will not change the name)"),
           Option( "-q", "int", 0, "charge of the molecule"),
           Option( "-scaleD", "float", 1.0, "additionally can scale halogen-sigmahole distance which is set by a rule (may be needed for stability)"),
           Option( "-sh", "bool", True, "if set to False, no sigmahole will be generated"),
           Option( "-clean", "bool", False, "clean working files"),
            ]

    help_text = ('OPLS sigma hole based on Jorgensen and Schyman, JCTC, 2012',
                     'Distances between halogen and sigma hole are fixed.',
                     'Charges for the sigma holes are fixed as well.',
                     'Only charges of halogen atoms are adjusted.',
                     '',
		     'CGenFF sigma hole: Gutierrez et al, Bioorg and Med Chem, 2016',
                     'Distances and sigma hole charges fixed.',
                     'The charges of halogen and first carbon atoms are changed (Table 1).',
		     '(NOTE: paramchem does not change charge of the first carbon)',
                     'LJ parameters and NBFIX are needed, but they are already',
                     'incorporated in the later versions of Charmm36m with CGenFF.',
                     '',
                     'GAFF sigma hole: Ibrahim, JCC, 2011',
                     'Distances between halogen and sigma hole are fixed.',
                     'Sigma hole charges set by approximating values observed for various ligands.',
                     'Only charges of halogen atoms are adjusted.',
                     'If Gaussian .log file is provided, proper RESP fitting is performed.',
                     'This way, sigma hole charges are fit and not pre-set.',
		     'If gaussian.log is provided, RESP fit will be performed',
                     '',
                     'The flag -rule allows mixing the rules to generate virtual particles.',
                     '"default" keeps the above described rules for the respective force-fields.',
                     'The rules may arbitrarily be used with any force-field by setting the -rule flag.',
                     ''
                  )

	# pass options, files and the command line to pymacs

    cmdl = Commandline( argv, options = options,fileoptions = files,program_desc = help_text, check_for_existing_files = False )

    ff = cmdl['-ff'].lower()
    if 'opls' in ff:
        ff = 'oplsaa'
    if ('charmm' in ff) or ('cgenff' in ff):
        ff = 'cgenff'
    if ('gaff2' in ff):
        ff = 'gaff2'
    elif ('gaff' in ff) or ('amber' in ff):
        ff = 'gaff'

    bSH = cmdl['-sh']

    bITPinp = cmdl.opt['-itp'].is_set
    bFFITPinp = cmdl.opt['-ffitp'].is_set
    itpfname = cmdl['-itp']
    ffitpfname = cmdl['-ffitp']

    ##########################
    ### identify the rules ###
    rule = cmdl['-rule'].lower()
    if rule.startswith('default'):
        if 'opls' in ff:
            rule = 'jorgensen'
        elif ('charmm' in ff) or ('cgenff' in ff):
            rule = 'gutierrez'
        elif 'gaff' in ff:
            rule = 'ibrahim'
    elif rule.startswith('jorgensen'):
        rule = 'jorgensen'
    elif rule.startswith('gutierrez'):
        rule = 'gutierrez'
    elif rule.startswith('ibrahim'):
        rule = 'ibrahim'
    else:
        sys.stdout.write('Rule set with the -rule flag is unknown. Exiting...\n')
        sys.exit(1)
    ########################### 

    if cmdl.opt['-log'].is_set==False and cmdl.opt['-esp'].is_set==False:
        if cmdl.opt['-pdb'].is_set==False:
            print('Need to provide Gaussian output or ESP file or PDB structure file')
            sys.exit(0)
        if cmdl.opt['-esp'].is_set==True and cmdl.opt['-xyz'].is_set==False and cmdl.opt['-pdb'].is_set==False:
            print('Need to provide XYZ or PBD when using ESP file')
            sys.exit(0)
    if cmdl.opt['-esp'].is_set==True and cmdl.opt['-xyz'].is_set==False and cmdl.opt['-pdb'].is_set==False:
        print('Need to provide XYZ or PDB when using ESP file')
        sys.exit(0)

    if cmdl.opt['-pdb'].is_set:
        # just read pdb
        if bITPinp or cmdl.opt['-log'].is_set or cmdl.opt['-esp'].is_set:
            m = Model().read(cmdl['-pdb'])
        # need to generate pdb first, then read it (only works for gaff now)
        elif( 'gaff' in ff):
            if cmdl.opt['-q'].is_set:
                run_acpype_from_pdb( cmdl['-pdb'],ff,charge=cmdl['-q'] )
            else:
                run_acpype_from_pdb( cmdl['-pdb'],ff )
            m = Model().read(cmdl['-pdb'])
            pdbname = cmdl['-pdb'].split('/')[-1].split('.')[-2]
            acpypepath = ''.join( cmdl['-pdb'].split('.')[0:-1] )+'.acpype'
            itpfname = '{0}/{1}_GMX.itp'.format(acpypepath,pdbname)
            ffitpfname = cmdl['-offitp']
            split_itp_ffitp( itpfname, ffitpfname, cmdl['-oitp'] )
            bITPinp = True
            bFFITPinp = True
            itpfname = cmdl['-oitp']
        else:
            print('For this ff cannot generate topology straight away from pdb')
            sys.exit(0)
#    sys.exit(0)

    scaleD = cmdl['-scaleD']
    bClean = cmdl['-clean']
    charge = cmdl['-q']

    bRESP = False
    if bITPinp:
        itp = TopolBase(itpfname)#, ff=ff)
    elif cmdl.opt['-log'].is_set:
        if 'gaff' not in ff:
            print('RESP fit can be done only for GAFF force field')
            sys.exit(0)
        gaussianLogFile = cmdl['-log']
        bRESP = True
    if bFFITPinp:
        ffitp = TopolBase( ffitpfname )#ff=ff )
    else:
        ffitp = TopolBase( filename=None )#ff=ff )

    if bITPinp:
        halogens = []
        halogens_nn = []
        if bSH==True:
            (halogens, halogens_nn) = find_halogens( itp,m=False )
        gen_sigmaholes( itp, m, ffitp, halogens, halogens_nn, rule, ff=ff, scaleD=scaleD )
#        write_ffitp( ffitp.atomtypes, itp, cmdl['-offitp'], ff=ff )
        write_ff( ffitp.atomtypes, cmdl['-offitp'], ff=ff )
    elif cmdl.opt['-log'].is_set or cmdl.opt['-esp'].is_set:
        randnum = random.randrange(1000,9999) # needed for temporary files
        # get charge
        if cmdl.opt['-log'].is_set:
            # get overall charge from gaussian output file
            charge = get_overall_charge_gaussian( cmdl['-log'] )
        elif cmdl.opt['-q'].is_set:
            charge = cmdl.opt['-q']
        else: # if nothing is provided, set charge to 0.0 
            charge = 0
        # get pdb
        if cmdl.opt['-log'].is_set:
            # extract pdb from Gaussian output
            fname_pdb = antechamber_pdb( cmdl['-log'], charge, randnum, ff=ff )
            m = Model().read(fname_pdb)
        elif cmdl.opt['-xyz'].is_set:
            m = structure_from_xyz( cmdl['-xyz'] )
        elif cmdl.opt['-pdb'].is_set:
            fname_pdb = cmdl['-pdb']
            m = Model().read(fname_pdb)
        else:
            print('Structure file (pdb or xyz) is needed when using -esp option')
            sys.exit(0)
        # find halogens
        (halogens, halogens_nn) = find_halogens( itp=False,m=m, bSH=bSH )
        # add sigmahole to pdb
        fname_pdb,sigmaholes = add_sigmahole_pdb( halogens, halogens_nn, m, randnum, rule, ff=ff, scaleD=scaleD )
        # create .ac file with sigmahole
        fname_ac = antechamber_ac( fname_pdb, m, halogens, sigmaholes, charge, randnum, ff )
        if cmdl.opt['-log'].is_set:
            # create esp with sigmahole
            fname_esp = generate_esp( cmdl['-log'], halogens, halogens_nn, m, randnum, rule, ff, scaleD=scaleD )
        else:
            fname_esp = generate_esp( cmdl['-esp'], halogens, halogens_nn, m, randnum, rule, ff, scaleD=scaleD, bESP=True )
#            fname_esp = cmdl['-esp']
        # perform resp fit
        fname_resp = generate_respfiles( fname_ac, fname_esp, randnum )
        # generate .ac with resp charges
        fname_ac_resp = antechamber_ac_resp( fname_ac, fname_resp, halogens, halogens_nn, sigmaholes, randnum, ff, charge )
        # generate frcmod
        fname_frcmod = generate_frcmod( fname_ac_resp, halogens, halogens_nn, randnum, rule, ff, scaleD=scaleD )
        # get mol2
        fname_mol2 = generate_mol2( fname_ac_resp, halogens, halogens_nn, sigmaholes, randnum, ff, charge )
        # run teLeap
        fname_prmtop,fname_inpcrd = run_teleap( fname_mol2, fname_frcmod, randnum, ff )
        # run acpype
        fname_top,fname_gro = run_acpype( fname_prmtop, fname_inpcrd, charge, randnum, ff )
        # top2itp
        itp = gen_itp( fname_top, fname_gro, randnum )
        # vsites and additional bonds for exclusions
        add_vsite_respfit( ff, itp, halogens, halogens_nn, sigmaholes, rule, scaleD=scaleD )
        # and additional bonds for exclusions
        add_sh_hnn_bond_respfit( ff, itp, halogens, halogens_nn, sigmaholes )
        # generate pairs as if sigma hole is the same atom as halogen (also remove unnecessary pairs)
        add_pairs_respfit( itp, halogens, sigmaholes )
        # remove any angles and dihedrals that might have been generated for the dummy vsites
        remove_angles_dihedrals( itp )
        # TODO: LJ radii (if RESP to be used with charmm)
        # write ffitp
        write_ff( itp.atomtypes, cmdl['-offitp'] )
        # remove atomtypes from .itp (they are already written to ffitp)
        itp.atomtypes = []
        # clean
        if bClean==True:
            cmd = "rm "+fname_pdb+" "+fname_ac+" "+fname_esp+" "+fname_resp+" "+fname_ac_resp+" "+fname_frcmod+" "+fname_mol2+" "+fname_prmtop+" "+fname_inpcrd+" "+fname_top+" ANTECHAMBER* ATOMTYPE* leap.in leap.log"
            os.system(cmd)
    else:
        print('Need to provide -itp or -log')
        sys.exit(0)
#        print halogens[0].name
#        print halogens_nn[0].name
#        sys.exit(0)

    if len(cmdl['-ligname'])>0:
        set_ligname( cmdl['-ligname'], m, itp )

    m.write(cmdl['-opdb'])
    itp.write( cmdl['-oitp'], stateBonded='A' )

main( sys.argv )



#!/usr/bin/env python
from pmx.model import Model
from pmx.forcefield import Topology
from pmx.alchemy import mutate, gen_hybrid_top
from pmx.gmx import set_gmxlib
from filecmp import cmp

# set GMXLIB variable
set_gmxlib()

def test_mutate(gf, tmpdir):
    # load, mutate, and save the PDB file
    m = Model(gf('alchemy/protein.pdb'), rename_atoms=True)
    m2 = mutate(m=m, mut_resid=6, mut_resname='F', ff='amber99sb-star-ildn-mut')
    # reference output file
    ref_output = gf('alchemy/mutant.pdb')
    # save mutant and compare
    outfile = str(tmpdir.mkdir("test_alchemy").join("mutant.pdb"))
    m2.write(outfile)
    cmp(ref_output, outfile)


def test_gen_hybrid_top(gf, tmpdir):
    # load topology, fill B states, then write a new topology file
    topol = Topology(gf('alchemy/topol.top'))
    pmxtop, _ = gen_hybrid_top(topol)
    # reference output file
    ref_output = gf('alchemy/pmxtop.top')
    # save top and compare
    outfile = str(tmpdir.mkdir("test_alchemy").join("pmxtop.top"))
    pmxtop.write(outfile)
    cmp(ref_output, outfile)

    # test case where one topology includes itps containing hybrid residues
    # ---------------------------------------------------------------------
    topol = Topology(gf('alchemy/inc_topol.top'))
    # fill B states for hybrid residues present in topol2.itp
    pmxtop, pmxitps = gen_hybrid_top(topol=topol, recursive=True)
    # reference output file
    ref_top = gf('alchemy/inc_pmxtop.top')
    ref_itp = gf('alchemy/inc_pmxitp_0.itp')
    # save top and compare
    out_top = str(tmpdir.join("test_alchemy/pmx.top"))
    out_itp = str(tmpdir.join("test_alchemy/pmx.itp"))
    pmxtop.write(out_top)
    pmxitps[0].write(out_itp)
    # compare
    cmp(ref_top, out_top)
    cmp(ref_itp, out_itp)



#!/usr/bin/env python

# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

"""Some functions to write parsers for different data formats.

Usage:

>>> lst = open(infile,'r').readlines()     # read file
>>> lst = kickOutComments(lst,'#')       # remove comments with # or ;
>>> lst = parseList('ifs',lst)            # parse each line and return
[integer, float, string] triples


>>> subl = readSection(lst,'[ begin ]','[ end ]') # get all lines between
[ begin ] and [ end ]
"""

from collections import OrderedDict


class ParserError(Exception):

    def __init__(self, s):
        self.s = s

    def __str__(self):
        return repr(self.s)


def kickOutComments(lines, comment='#'):
    ret = []
    for line in lines:
        if comment in line:
            idx = line.index(comment)
            new_line = line[:idx].strip()
            if new_line:
                ret.append(new_line)
        else:
            new_line = line.strip()
            if new_line:
                ret.append(new_line)
    return ret


def readSection(lines, begin, end):
    ret = []
    for i, line in enumerate(lines):
        if line.strip() == begin:
            for line2 in lines[i+1:]:
                if end not in line2:
                    ret.append(line2)
                else:
                    return ret
    return ret


def __parse_error(msg, line):
    s = "pmx_Error_> %s" % msg
    s += "\nTrouble is here -> %s" % line
    raise ParserError(s)


def __parse_entry(entr, tp):
    new = None
    if tp == 's':
        new = entr
    elif tp == 'i':
        try:
            new = int(entr)
        except:
            __parse_error("Integer conversion failed", entr)
    elif tp == 'f':
        try:
            new = float(entr)
        except:
            __parse_error("Float conversion failed", entr)
    return new


def parseList(format_string, lst, ignore_missing=False):
    ret = []
    format_list = [a for a in format_string]
    for line in lst:
        entr = line.split()
        if len(entr) != len(format_list):
            if not ignore_missing or len(entr) < len(format_list):
                __parse_error("Cannot convert line into format: %s" %
                              format_string, line)
        new_list = []
        for i, tp in enumerate(format_list):
            new_list.append(__parse_entry(entr[i], tp))
        ret.append(new_list)
    return ret


def read_and_format(filename, format_string, comment='#',
                    ignore_missing=False):
    """
    format_string : str
        e.g. 'is' = [int, str]; 'iis' = [int, int, str] etc.
    """
    l = open(filename).readlines()
    if comment is not None:
        l = kickOutComments(l, comment)
    n = parseList(format_string, l, ignore_missing)
    return n


# ------------------------------------------
# some file format parsers frequently needed
# ------------------------------------------
def read_fasta(fn):
    l = open(fn).readlines()
    l = kickOutComments(l)
    dic = OrderedDict()
    name = None
    for line in l:
        if line.startswith('>'):
            if name is not None:
                dic[name] = new_s
            new_s = ''
            name = line[1:].strip()
        else:
            seq = line.strip().upper()
            new_s += seq
    dic[name] = new_s
    return dic


def read_xvg(fn,  style='xy'):
    l = open(fn).readlines()
    l = kickOutComments(l, '@')
    l = kickOutComments(l, '#')
    l = kickOutComments(l, '&')
    res = parseList('ff', l)
    if style == 'list':
        return res
    else:
        x = list(map(lambda a: a[0],  res))
        y = list(map(lambda a: a[1],  res))
        return x,  y



"""pmx: a toolkit for free-energy calculation setup/analysis and
biomolecular structure handling.

For installation type the command::
  python setup.py install
or
  pip install .
"""

from setuptools import setup, Extension
import versioneer


def readme():
    with open('README.rst') as f:
        return f.read()

# ----------
# Extensions
# ----------
pmx = Extension('pmx._pmx',
                libraries=['m'],
                include_dirs=['src/pmx/extensions/pmx'],
                sources=['src/pmx/extensions/pmx/Geometry.c',
                         'src/pmx/extensions/pmx/wrap_Geometry.c',
                         'src/pmx/extensions/pmx/init.c',
                         'src/pmx/extensions/pmx/Energy.c']
                )

xdrio = Extension('pmx._xdrio',
                  libraries=['m'],
                  include_dirs=['src/pmx/extensions/xdr'],
                  sources=['src/pmx/extensions/xdr/xdrfile.c',
                           'src/pmx/extensions/xdr/xdrfile_trr.c',
                           'src/pmx/extensions/xdr/xdrfile_xtc.c']
                  )
extensions = [pmx, xdrio]

# -----
# Setup
# -----
setup(name='pmx_biobb',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      description='Toolkit for free-energy calculation setup/analysis and biomolecular structure handling',
      long_description_content_type='text/x-rst',
      long_description=readme(),
      classifiers=[
          'Programming Language :: Python :: 3.10',
          'Intended Audience :: Science/Research',
          'Topic :: Scientific/Engineering',
      ],
      url='https://github.com/deGrootLab/pmx/tree/develop',
      author='Daniel Seeliger, Vytautas Gapsys',
      author_email='d.seeliger@gmx.net, vytautas.gapsys@gmail.com',
      license='LGPL-3.0',
      packages=['pmx'],
      package_dir={'': 'src'},
      include_package_data=True,
      zip_safe=False,
      ext_modules=extensions,
      tests_require=['pytest'],
      install_requires=['numpy', 'scipy', 'matplotlib', 'rdkit'],
      python_requires=">=3.10",
      entry_points={'console_scripts': ['pmx = pmx.scripts.cli:entry_point']},
      )



#!/usr/bin/env python

# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

"""This module comtains some functions to build protein
structures from scratch.

Most important:
- build_chain

Usage:
    >>> from pmx.builder import *
    >>> ch = build_chain('AAAAAA')    # builds an ordinary polyalanine
    >>> ch.write('polyala.pdb')       # store file
    >>> ch = build_chain('FGHRTCV',hydrogens = False) # build chain without H
    >>> ch = build_chain('FGHRTCV',ss='HHHHHHH') build as helix
    >>> ch = build_chain('FGHRTCV',dihedrals = ((phi1,psi1,omega1),(...)))
    build chain with defined dihedral angles
"""

import sys
import copy as cp
from numpy import array, pi, linalg, cross
from . import library
from .geometry import Rotation
from .chain import Chain
from .atom import Atom
from .model import Model
from .molecule import Molecule
from . import geometry


def add_bp(m, strand=None, bRNA=False):
    if strand:
        N = len(strand)/2
        if bRNA:
            N = len(strand)
    else:
        N = 1

    r = Rotation([0, 0, 0], [0, 0, 1])
    phi = 0.0
    phi = N*36*pi/180.
    for atom in m.atoms:
        atom.x = r.apply(atom.x, phi)
    for atom in m.atoms:
        atom.x[2] += N*3.4


def make_3ter(r):
    # we add a proton at O3'
    c3, o3 = r.fetchm(['C3\'', 'O3\''])
    v = array(o3.x)-array(c3.x)
    normed = v*1./linalg.norm(v)
    newpos = array(o3.x) + normed
    a = Atom(name='H3T', x=newpos)
    r.append(a)


def make_5ter(r):
    del r['O1P']
    del r['O2P']
    # we add a proton at O5'
    a = r.fetch_atoms('P')[0]
    a.name = 'H5T'
    h5t, o5 = r.fetchm(['H5T', 'O5\''])
    v = array(h5t.x)-array(o5.x)
    normed = v*1./linalg.norm(v)
    newpos = array(o5.x) + normed
    h5t.x = newpos


def build_dna_strand(seq):

    dic = library.pmx_data_file('bp.pkl')

    seq = seq.lower()
    ss = []
    for x in seq:
        ss.append(x)
    new = []
    new.extend(dic[ss[0]].residues)
    ss.pop(0)
    while ss:
        newbp = ss.pop(0)
        newm = dic[newbp].copy()
        add_bp(newm, new)
        new.extend(newm.residues)
    chA = []
    chB = []
    a = 1
    b = 1
    for r in new:
        if r.chain_id == 'B':
            r.set_resid(a)
            a += 1
            chA.append(r)
        else:
            chB.append(r)
    chB.reverse()
    for r in chB:
        r.set_resid(b)
        b += 1
    mm = Model(residues=chA+chB)
    mm.chains[1].set_chain_id('C')
    mm.chains[0].set_chain_id('A')
    mm.chains[1].set_chain_id('B')
    for chain in mm.chains:
        r = chain.residues[0]
        r.set_resname(r.resname+'5')
        make_5ter(r)
        r = chain.residues[-1]
        r.set_resname(r.resname+'3')
        make_3ter(r)
    return mm


def build_rna_strand(seq):

    dic = library.pmx_data_file('rna.pkl')

    seq = seq.lower()
    ss = []
    for x in seq:
        ss.append(x)
    new = []
    new.extend(dic[ss[0]].residues)
    ss.pop(0)
    while ss:
        newbp = ss.pop(0)
        newm = dic[newbp].copy()
        add_bp(newm, new, True)
        new.extend(newm.residues)
    chA = []
    a = 1
    for r in new:
        r.set_resid(a)
        a += 1
        chA.append(r)
    mm = Model(residues=chA)
    mm.chains[0].set_chain_id('A')
    for chain in mm.chains:
        r = chain.residues[0]
        r.set_resname(r.resname+'5')
        make_5ter(r)
        r = chain.residues[-1]
        r.set_resname(r.resname+'3')
        make_3ter(r)
    return mm


def get_fragments():
    dic = library.pmx_data_file('fragments.pkl')
    n = len(dic.keys())
    print("pmx__> # Fragments loaded: %d" % n, file=sys.stderr)
    return dic


def read_pdb_with_connect(f):
    m = Model().read(f)
    m.nm2a()
    mol = m.residues[0]
    l = open(f).readlines()
    conn = []
    for line in l:
        if line[0:6] == 'CONECT':
            conn.append([int(x) for x in line[6:].split()])
    for lst in conn:
        if len(lst) > 1:
            atoms = mol.get_by_id(lst)
            a0 = atoms[0]
            for atom in atoms[1:]:
                a0.bonds.append(atom)
    return mol


def write_pdb_with_connect(mol, f, n=1):
    if not hasattr(f, "write"):
        fp = open(f, "w")
    else:
        fp = f
    print("MODEL%5d" % n, file=fp)
    for atom in mol.atoms:
        print('{}'.format(atom), file=fp)
    for atom in mol.atoms:
        s = "CONECT%5d" % atom.id
        for a in atom.bonds:
            s += '%5d' % a.id
        print('{}'.format(s), file=fp)


def attach_group(atom, mol):
    master = atom.molecule

    bb = atom.bonds[0]
    R = mol.fetch_atoms('R#')[0]
    bR = R.bonds[0]
    diff = array(R.x)-array(bb.x)
    if bR.name[0] == 'C':
        l = 1.54
    elif bR.name[0] == 'S':
        l = 1.8
    elif bR.name[0] == 'O':
        l = 1.4
    else:
        l = 1.5
    for a in mol.atoms:
        a.x = array(a.x)-diff
    ang = R.angle(bR, atom)
    v1 = array(atom.x)-array(bb.x)
    v2 = array(bR.x)-array(R.x)
    cr = cross(v1, v2)
    x = 1./linalg.norm(cr)
    cr = cr*x
    cr = R.x + cr
    r = Rotation(cr, R.x)
    for a in mol.atoms:
        a.x = r.apply(a.x, ang)
    vec = array(bb.x) - array(bR.x)
    x = 1./linalg.norm(vec)
    for a in mol.atoms:
        a.x += vec*(1-x*l)
    for a in master.atoms:
        if atom in a.bonds:
            a.bonds.remove(atom)
    master.remove_atom(atom)
    bR.bonds.remove(R)
    mol.remove_atom(R)
    bR.bonds.append(bb)
    bb.bonds.append(bR)
    for a in mol.atoms:
        master.append(a)


def _is_mainchain( a ):
    if a.name=='N':
        return(True)
    if a.name=='H' or a.name=='HN' or a.name=='H1' or a.name=='H2' or a.name=='H3':
        return(True)
    if a.name=='CA':
        return(True)
    if a.name=='C':
        return(True)
    if a.name=='O' or a.name=='OXT' or a.name=='OX' or a.name=='OT':
        return(True)
    return(False)

def make_stereoisomer_residue( m ):
    mtmp = cp.deepcopy( m )
    # fit
    N1, CA1, C1 = m.fetchm(['N', 'CA', 'C'])
    N2, CA2, C2 = mtmp.fetchm(['N', 'CA', 'C'])
    geometry.fit_atoms( [N1,CA1,C1], [C2,CA2,N2], mtmp.atoms )
    for atmp,a in zip(mtmp.atoms,m.atoms):
        if _is_mainchain( a )==False:
            a.x = atmp.x
    return(m)

def make_residue(key, hydrogens=True, chirality='L'):
    """ returns a molecule object with
    default geometry"""

    if key not in library._aacids:
        raise(KeyError, "Residue %s not known" % key)
    m = Molecule()
    m.unity = 'A'
    m.resname = key
    for i, entry in enumerate(library._aacids[key]):
        if hydrogens is False and entry[0][0] == 'H':
            continue
        else:
            a = Atom()
            a.id = i+1
            a.name = entry[0]
            a.symbol = a.name[0]
            a.x = entry[1]
            a.occ = 1.
            a.resname = key
            a.m = library._atommass[a.symbol]
            a.unity = 'A'
            m.atoms.append(a)

    if chirality=='D':
        m = make_stereoisomer_residue( m )
    return m


def build_chain(sequence, dihedrals=None,
                hydrogens=True, ss=None, chain_id='X', bCyclic=False, bCap=False,
                chirality=None):

    # chirality 
    if chirality==None:
        chirality = ''
        for s in sequence:
            chirality = chirality+'L'

    # if cyclic, form dihedrals
    n = len(sequence)
    if bCyclic==True:
        dihedrals = []
        fudge = float(n)/5.0 # empirical fudge to have some space between the termini
        dphi = 2.0*360.0/float(n+fudge)
        if bCap==True: # leave space for caps
            dphi = 2.0*360.0/float(n+fudge+1)
        dpsi = 0.0
        dom = 180.0
        sign = -1.0
        for i in range(0,n):
            dihedrals.append( (sign*dphi,-1.0*sign*dpsi,dom) )
            sign = -1.0*sign

    ch = Chain()
    if dihedrals is None:
        if ss is None:
            dihedrals = []
            for i in range(len(sequence)):
                dihedrals.append(library._extended)
        else:
            dihedrals = []
            for i, s in enumerate(ss):
                if s == '-':
                    dihedrals.append(library._extended)
                elif s == 'H':
                    dihedrals.append(library._helix)
                elif s == 'E':
                    dihedrals.append(library._anti_beta)

    rl = []
    start = make_residue(library._aacids_dic[sequence[0]], hydrogens=hydrogens, chirality=chirality[0])
    start.set_resid(1)
    set_psi(start, dihedrals[0][1])
    rl.append(start)
    for i in range(1,n):
        aa = sequence[i]
        ld = chirality[i]
        phi, psi, om = dihedrals[i]
        phi2, psi2, om2 = dihedrals[i-1]
        mol = attach_aminoacid(start, library._aacids_ext_amber[aa],
                               hydrogens=hydrogens,
                               phi=phi, psi=psi, omega=om2,chirality=ld)
        mol.set_resid(i+1)
        set_phi(start, mol, dihedrals[i][0])
        start = mol
        rl.append(mol)
    al = []
    for r in rl:
        al += r.atoms
    ch.unity = 'A'
    ch.atoms = al
    ch.renumber_atoms()
    ch.residues = rl
    ch.set_chain_id(chain_id)
    for atom in ch.atoms:
        atom.unity = 'A'

    if bCap==True:
        ch.add_nterm_cap()
        ch.add_cterm_cap()

    return ch


def set_phi(mol1, mol2, phi):
    # select c, n, ca, c
    C = mol1.fetchm(['C'])[0]
    N, CA, C2 = mol2.fetchm(['N', 'CA', 'C'])
    dih = C.dihedral(N, CA, C2)
    phi = pi/180*phi
    delta = phi-dih

    # rotation is around N -> CA
    r = Rotation(N.x, CA.x)

    for atom in mol2.atoms:
        if atom.name not in ['N', 'H']:
            atom.x = r.apply(atom.x, delta)


def set_psi(mol, phi):

    # select n, c, ca, o
    N, CA, C, O = mol.fetchm(['N', 'CA', 'C', 'O'])
    dih = N.dihedral(CA, C, O)  # -pi
    phi = pi/180*phi - pi  # psi is defined with N'
    delta = phi - dih
    r = Rotation(CA.x, C.x)

    # only O is affected
    O.x = r.apply(O.x, delta)


def set_omega(mol1, mol2, phi):

    # select c, n, ca, c
    CA, C = mol1.fetchm(['CA', 'C'])
    N, CA2, C2 = mol2.fetchm(['N', 'CA', 'C'])
    dih = CA.dihedral(C, N, CA2)  # -pi
    phi = pi/180*phi
    delta = phi-dih

    # rotation is around C -> N
    r = Rotation(C.x, N.x)
    for atom in mol2.atoms:
        if atom.name != 'N':
            atom.x = r.apply(atom.x, delta)


def attach_aminoacid(mol, resname, hydrogens=True,
                     omega=180., phi=-139., psi=135., chirality='L'):

    new = make_residue(resname, hydrogens=hydrogens, chirality=chirality)
    # we need Ca, C and O from mol
    Ca, C, O = mol.fetchm(['CA', 'C', 'O'])
    # and N from new
    N, CA2 = new.fetchm(['N', 'CA'])

    l = 1.33  # length of peptide bond

    v = array(C.x)-array(Ca.x)
    x = 1./linalg.norm(v)
    vec = v*x*l
    newpos = array(C.x)+vec
    t = array(N.x)-newpos

    # shift to new position
    for a in new.atoms:
        a.x = array(a.x)-t

    # gen rotation vector
    # this rotation yields the correct
    # O-C-N angle
    v1 = array(C.x)-array(Ca.x)
    v2 = array(O.x)-array(C.x)
    cr = cross(v1, v2)
    x = 1./linalg.norm(cr)
    cr = cr*x
    cr = C.x + cr

    # do first rotation
    r = Rotation(cr, C.x)
    for atom in new.atoms:
        atom.x = r.apply(atom.x, pi/3.)

    # do next rotation
    # here we correct the C-N-Ca angle
    v1 = array(N.x)-array(C.x)
    v2 = array(CA2.x)-array(N.x)
    cr = cross(v1, v2)
    x = 1./linalg.norm(cr)
    cr = cr*x
    cr = N.x + cr
    an = N.angle(C, CA2)
    delta = an - 120*pi/180
    # do second rotation
    r = Rotation(cr, N.x)
    for atom in new.atoms:
        if atom.name != 'N':
            atom.x = r.apply(atom.x, -delta)

    # correct H if we have one
    H = new.fetch('H')
    if H:
        H = H[0]
        # we place the H-atom along the
        # N-Ca vector.
        # then we move it
        v1 = array(CA2.x)-array(N.x)
        x = 1./linalg.norm(v1)
        H.x = N.x + v1*x
        H.x = r.apply(H.x, 2*pi/3.)

    set_omega(mol, new, omega)
    set_phi(mol, new, phi)
    set_psi(new, psi)

    return new
